#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 02-01-2021
           """

from typing import Union

import numpy
from numpy.random.mtrand import RandomState

__all__ = ["get_sampler"]


def get_sampler(seed: Union[RandomState, int] = None) -> RandomState:
    """ """
    if isinstance(seed, RandomState):
        return seed
    return numpy.random.RandomState(seed)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 03-05-2021
           """

from enum import Enum
from typing import Sequence, Tuple, Union

import numpy
from scipy.spatial import distance

__all__ = [
    "mouth_aspect_ratio",
    "Dlib68faciallandmarksindices",
    "Dlib5faciallandmarksindices",
    "rect_to_bounding_box",
    "shape_to_ndarray",
    "eye_aspect_ratio",
]


class ExtendTuple(Tuple):
    def __or__(self, *other) -> "ExtendTuple":
        return ExtendTuple((*self, *other))

    def __and__(self, *other) -> "ExtendTuple":
        return ExtendTuple((*self, *other))


class Dlib68faciallandmarksindices(Enum):
    mouth = (48, 67 + 1)
    inner_mouth = (60, 67 + 1)
    right_eyebrow = (17, 21 + 1)
    left_eyebrow = (22, 26 + 1)
    right_eye = (36, 41 + 1)
    left_eye = (42, 47 + 1)
    nose = (27, 35 + 1)
    jaw = (0, 16 + 1)

    @staticmethod
    def slice(
        seq: Sequence,
        ind: Union[
            "Dlib68faciallandmarksindices", Tuple["Dlib68faciallandmarksindices"]
        ],
    ):
        """

        :param seq:
        :param ind:
        :return:
        """
        if isinstance(ind, Tuple):
            agg = []
            for (
                i
            ) in (
                ind
            ):  # Some flag implementation would probably be faster and more valid.
                agg.extend(Dlib68faciallandmarksindices.slice(seq, i))
            return agg
        start, end = ind.value
        return seq[start:end]

    def __or__(
        self, other: "Dlib68faciallandmarksindices"
    ) -> Tuple["Dlib68faciallandmarksindices"]:
        return ExtendTuple().__or__(self, other)

    def __and__(
        self, other: "Dlib68faciallandmarksindices"
    ) -> Tuple["Dlib68faciallandmarksindices"]:
        return ExtendTuple().__and__(self, other)


class Dlib5faciallandmarksindices(Enum):
    """
    dlib_utilities’s 5-point facial landmark detector
    """

    right_eye = (2, 3 + 1)
    left_eye = (0, 1 + 1)
    nose = (4, 4 + 1)

    @staticmethod
    def slice(
        seq: Sequence,
        ind: Union["Dlib5faciallandmarksindices", Tuple["Dlib5faciallandmarksindices"]],
    ):
        """

        :param seq:
        :param ind:
        :return:
        """
        if isinstance(ind, Tuple):
            agg = []
            for (
                i
            ) in (
                ind
            ):  # Some flag implementation would probably be faster and more valid.
                agg.extend(Dlib5faciallandmarksindices.slice(seq, i))
            return agg
        start, end = ind.value
        return seq[start:end]

    def __or__(
        self, other: "Dlib5faciallandmarksindices"
    ) -> Tuple["Dlib5faciallandmarksindices"]:
        return ExtendTuple().__or__(self, other)

    def __and__(
        self, other: "Dlib5faciallandmarksindices"
    ) -> Tuple["Dlib5faciallandmarksindices"]:
        return ExtendTuple().__and__(self, other)


def rect_to_bounding_box(rect) -> Tuple[float, float, float, float]:
    """
    # take a bounding predicted by dlib_utilities and convert it
    # to the format (x, y, w, h)"""
    x = rect.left()
    y = rect.top()

    return x, y, rect.right() - x, rect.bottom() - y  # return a tuple of (x, y, w, h)


def shape_to_ndarray(shape, dtype: str = "int"):
    """

    :param shape:
    :param dtype:
    :return:
    """
    coordinates = numpy.zeros(
        (shape.num_parts, 2), dtype=dtype
    )  # initialize the list of (x, y)-coordinates

    # loop over all facial landmarks and convert them
    # to a 2-tuple of (x, y)-coordinates
    for i in range(0, shape.num_parts):
        coordinates[i] = (shape.part(i).x, shape.part(i).y)

    return coordinates  # return the list of (x, y)-coordinates


def mouth_aspect_ratio(coordinates: Sequence[Sequence]) -> float:
    """

    :param coordinates:
    :return:"""
    average = (
        distance.euclidean(coordinates[3], coordinates[9])
        + distance.euclidean(coordinates[2], coordinates[10])
        + distance.euclidean(coordinates[4], coordinates[8])
    ) / 3
    return average / distance.euclidean(coordinates[0], coordinates[6])


def eye_aspect_ratio(coordinates: Sequence[Sequence]) -> float:
    """

    :param coordinates:
    :return:
    """
    # compute the euclidean distances between the two sets of
    # vertical eye landmarks (x, y)-coordinates
    A = distance.euclidean(coordinates[1], coordinates[5])
    B = distance.euclidean(coordinates[2], coordinates[4])

    # compute the euclidean distance between the horizontal
    # eye landmark (x, y)-coordinates
    C = distance.euclidean(coordinates[0], coordinates[3])

    # compute the eye aspect ratio
    return (A + B) / (2.0 * C)


# To improve our blink detector, Soukupová and Čech recommend constructing a 13-dim feature vector of eye aspect ratios (N-th frame, N – 6 frames, and N + 6 frames), followed by feeding this feature vector into a Linear SVM for classification.

if __name__ == "__main__":

    def asud():
        a = list(range(99))
        slices = (
            Dlib68faciallandmarksindices.left_eye
            | Dlib68faciallandmarksindices.right_eye
            | Dlib68faciallandmarksindices.nose & Dlib68faciallandmarksindices.mouth
        )
        print(Dlib68faciallandmarksindices.slice(a, slices))

    def as34ud():
        a = list(range(99))
        print(
            Dlib68faciallandmarksindices.slice(a, Dlib68faciallandmarksindices.left_eye)
        )

    asud()
    as34ud()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


__author__ = "heider"
__doc__ = r"""

           Created on 01/02/2022
           """

__all__ = ["color"]

import re
from functools import partial

"""
Map of CSS color names to RGB integer values.
"""
css_colors = {
    "aliceblue": (240, 248, 255),
    "antiquewhite": (250, 235, 215),
    "aqua": (0, 255, 255),
    "aquamarine": (127, 255, 212),
    "azure": (240, 255, 255),
    "beige": (245, 245, 220),
    "bisque": (255, 228, 196),
    "black": (0, 0, 0),
    "blanchedalmond": (255, 235, 205),
    "blue": (0, 0, 255),
    "blueviolet": (138, 43, 226),
    "brown": (165, 42, 42),
    "burlywood": (222, 184, 135),
    "cadetblue": (95, 158, 160),
    "chartreuse": (127, 255, 0),
    "chocolate": (210, 105, 30),
    "coral": (255, 127, 80),
    "cornflowerblue": (100, 149, 237),
    "cornsilk": (255, 248, 220),
    "crimson": (220, 20, 60),
    "cyan": (0, 255, 255),
    "darkblue": (0, 0, 139),
    "darkcyan": (0, 139, 139),
    "darkgoldenrod": (184, 134, 11),
    "darkgray": (169, 169, 169),
    "darkgreen": (0, 100, 0),
    "darkgrey": (169, 169, 169),
    "darkkhaki": (189, 183, 107),
    "darkmagenta": (139, 0, 139),
    "darkolivegreen": (85, 107, 47),
    "darkorange": (255, 140, 0),
    "darkorchid": (153, 50, 204),
    "darkred": (139, 0, 0),
    "darksalmon": (233, 150, 122),
    "darkseagreen": (143, 188, 143),
    "darkslateblue": (72, 61, 139),
    "darkslategray": (47, 79, 79),
    "darkslategrey": (47, 79, 79),
    "darkturquoise": (0, 206, 209),
    "darkviolet": (148, 0, 211),
    "deeppink": (255, 20, 147),
    "deepskyblue": (0, 191, 255),
    "dimgray": (105, 105, 105),
    "dimgrey": (105, 105, 105),
    "dodgerblue": (30, 144, 255),
    "firebrick": (178, 34, 34),
    "floralwhite": (255, 250, 240),
    "forestgreen": (34, 139, 34),
    "fuchsia": (255, 0, 255),
    "gainsboro": (220, 220, 220),
    "ghostwhite": (248, 248, 255),
    "gold": (255, 215, 0),
    "goldenrod": (218, 165, 32),
    "gray": (128, 128, 128),
    "green": (0, 128, 0),
    "greenyellow": (173, 255, 47),
    "grey": (128, 128, 128),
    "honeydew": (240, 255, 240),
    "hotpink": (255, 105, 180),
    "indianred": (205, 92, 92),
    "indigo": (75, 0, 130),
    "ivory": (255, 255, 240),
    "khaki": (240, 230, 140),
    "lavender": (230, 230, 250),
    "lavenderblush": (255, 240, 245),
    "lawngreen": (124, 252, 0),
    "lemonchiffon": (255, 250, 205),
    "lightblue": (173, 216, 230),
    "lightcoral": (240, 128, 128),
    "lightcyan": (224, 255, 255),
    "lightgoldenrodyellow": (250, 250, 210),
    "lightgray": (211, 211, 211),
    "lightgreen": (144, 238, 144),
    "lightgrey": (211, 211, 211),
    "lightpink": (255, 182, 193),
    "lightsalmon": (255, 160, 122),
    "lightseagreen": (32, 178, 170),
    "lightskyblue": (135, 206, 250),
    "lightslategray": (119, 136, 153),
    "lightslategrey": (119, 136, 153),
    "lightsteelblue": (176, 196, 222),
    "lightyellow": (255, 255, 224),
    "lime": (0, 255, 0),
    "limegreen": (50, 205, 50),
    "linen": (250, 240, 230),
    "magenta": (255, 0, 255),
    "maroon": (128, 0, 0),
    "mediumaquamarine": (102, 205, 170),
    "mediumblue": (0, 0, 205),
    "mediumorchid": (186, 85, 211),
    "mediumpurple": (147, 112, 219),
    "mediumseagreen": (60, 179, 113),
    "mediumslateblue": (123, 104, 238),
    "mediumspringgreen": (0, 250, 154),
    "mediumturquoise": (72, 209, 204),
    "mediumvioletred": (199, 21, 133),
    "midnightblue": (25, 25, 112),
    "mintcream": (245, 255, 250),
    "mistyrose": (255, 228, 225),
    "moccasin": (255, 228, 181),
    "navajowhite": (255, 222, 173),
    "navy": (0, 0, 128),
    "oldlace": (253, 245, 230),
    "olive": (128, 128, 0),
    "olivedrab": (107, 142, 35),
    "orange": (255, 165, 0),
    "orangered": (255, 69, 0),
    "orchid": (218, 112, 214),
    "palegoldenrod": (238, 232, 170),
    "palegreen": (152, 251, 152),
    "paleturquoise": (175, 238, 238),
    "palevioletred": (219, 112, 147),
    "papayawhip": (255, 239, 213),
    "peachpuff": (255, 218, 185),
    "peru": (205, 133, 63),
    "pink": (255, 192, 203),
    "plum": (221, 160, 221),
    "powderblue": (176, 224, 230),
    "purple": (128, 0, 128),
    "rebeccapurple": (102, 51, 153),
    "red": (255, 0, 0),
    "rosybrown": (188, 143, 143),
    "royalblue": (65, 105, 225),
    "saddlebrown": (139, 69, 19),
    "salmon": (250, 128, 114),
    "sandybrown": (244, 164, 96),
    "seagreen": (46, 139, 87),
    "seashell": (255, 245, 238),
    "sienna": (160, 82, 45),
    "silver": (192, 192, 192),
    "skyblue": (135, 206, 235),
    "slateblue": (106, 90, 205),
    "slategray": (112, 128, 144),
    "slategrey": (112, 128, 144),
    "snow": (255, 250, 250),
    "springgreen": (0, 255, 127),
    "steelblue": (70, 130, 180),
    "tan": (210, 180, 140),
    "teal": (0, 128, 128),
    "thistle": (216, 191, 216),
    "tomato": (255, 99, 71),
    "turquoise": (64, 224, 208),
    "violet": (238, 130, 238),
    "wheat": (245, 222, 179),
    "white": (255, 255, 255),
    "whitesmoke": (245, 245, 245),
    "yellow": (255, 255, 0),
    "yellowgreen": (154, 205, 50),
}

# ANSI color names. There is also a "default"
COLORS = ("black", "red", "green", "yellow", "blue", "magenta", "cyan", "white")

# ANSI style names
STYLES = (
    "none",
    "bold",
    "faint",
    "italic",
    "underline",
    "blink",
    "blink2",
    "negative",
    "concealed",
    "crossed",
)


def parse_rgb(s):
    if not isinstance(s, str):
        raise ValueError("Could not parse color '{0}'".format(s))
    s = s.strip().replace(" ", "").lower()
    # simple lookup
    rgb = css_colors.get(s)
    if rgb is not None:
        return rgb

    # 6-digit hex
    match = re.match("#([a-f0-9]{6})$", s)
    if match:
        core = match.group(1)
        return tuple(int(core[i : i + 2], 16) for i in range(0, 6, 2))

    # 3-digit hex
    match = re.match("#([a-f0-9]{3})$", s)
    if match:
        return tuple(int(c * 2, 16) for c in match.group(1))

    # rgb(x,y,z)
    match = re.match(r"rgb\((\d+,\d+,\d+)\)", s)
    if match:
        return tuple(int(v) for v in match.group(1).split(","))

    raise ValueError("Could not parse color '{0}'".format(s))


def is_string(obj):
    """
    Is the given object a string?
    """
    return isinstance(obj, str)


def _join(*values):
    """
    Join a series of values with semicolons. The values
    are either integers or strings, so stringify each for
    good measure. Worth breaking out as its own function
    because semicolon-joined lists are core to ANSI coding.
    """
    return ";".join(str(v) for v in values)


def _color_code(spec, base):
    """
    Workhorse of encoding a color. Give preference to named colors from
    ANSI, then to specific numeric or tuple specs. If those don't work,
    try looking up CSS color names or parsing CSS color specifications
    (hex or rgb).

    :param str|int|tuple|list spec: Unparsed color specification
    :param int base: Either 30 or 40, signifying the base value
        for color encoding (foreground and background respectively).
        Low values are added directly to the base. Higher values use `
        base + 8` (i.e. 38 or 48) then extended codes.
    :returns: Discovered ANSI color encoding.
    :rtype: str
    :raises: ValueError if cannot parse the color spec.
    """
    if is_string(spec):
        spec = spec.strip().lower()

    if spec == "default":
        return _join(base + 9)
    elif spec in COLORS:
        return _join(base + COLORS.index(spec))
    elif isinstance(spec, int) and 0 <= spec <= 255:
        return _join(base + 8, 5, spec)
    elif isinstance(spec, (tuple, list)):
        return _join(base + 8, 2, _join(*spec))
    else:
        rgb = parse_rgb(spec)
        # parse_rgb raises ValueError if cannot parse spec
        # or returns an rgb tuple if it can
        return _join(base + 8, 2, _join(*rgb))


def color(s, fg=None, bg=None, style=None):
    """
    Add ANSI colors and styles to a string.

    :param str s: String to format.
    :param str|int|tuple fg: Foreground color specification.
    :param str|int|tuple bg: Background color specification.
    :param str: Style names, separated by '+'
    :returns: Formatted string.
    :rtype: str (or unicode in Python 2, if s is unicode)
    """
    codes = []

    if fg:
        codes.append(_color_code(fg, 30))
    if bg:
        codes.append(_color_code(bg, 40))
    if style:
        for style_part in style.split("+"):
            if style_part in STYLES:
                codes.append(STYLES.index(style_part))
            else:
                raise ValueError('Invalid style "%s"' % style_part)

    if codes:
        template = "\x1b[{0}m{1}\x1b[0m"
        return template.format(_join(*codes), s)
    else:
        return s


def strip_color(s):
    """
    Remove ANSI color/style sequences from a string. The set of all possible
    ANSI sequences is large, so does not try to strip every possible one. But
    does strip some outliers seen not just in text generated by this module, but
    by other ANSI colorizers in the wild. Those include `\x1b[K` (aka EL or
    erase to end of line) and `\x1b[m`, a terse version of the more common
    `\x1b[0m`.
    """
    return re.sub("\x1b\\[(K|.*?m)", "", s)


def ansilen(s):
    """
    Given a string with embedded ANSI codes, what would its
    length be without those codes?
    """
    return len(strip_color(s))


# Foreground color shortcuts
black = partial(color, fg="black")
red = partial(color, fg="red")
green = partial(color, fg="green")
yellow = partial(color, fg="yellow")
blue = partial(color, fg="blue")
magenta = partial(color, fg="magenta")
cyan = partial(color, fg="cyan")
white = partial(color, fg="white")

# Style shortcuts
bold = partial(color, style="bold")
none = partial(color, style="none")
faint = partial(color, style="faint")
italic = partial(color, style="italic")
underline = partial(color, style="underline")
blink = partial(color, style="blink")
blink2 = partial(color, style="blink2")
negative = partial(color, style="negative")
concealed = partial(color, style="concealed")
crossed = partial(color, style="crossed")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from abc import ABC, abstractmethod
from typing import Mapping, Sequence

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 09/10/2019
           """
__all__ = ["LineWriterMixin"]


class LineWriterMixin(ABC):
    """
    Writer mixin that provides an interface for 'writing' line charts"""

    @abstractmethod
    def line(
        self,
        tag: str,
        values: list,
        step: int,
        x_labels: Sequence = None,
        y_label: str = "Magnitude",
        x_label: str = "Sequence",
        plot_kws: Mapping = None,  # Separate as parameters name collisions might occur
        **kwargs
    ) -> None:
        """

            :param values:
            :param x_labels:
            :param y_label:
            :param x_label:
            :param plot_kws:
        :param tag:
        :type tag:
        :param step:
        :type step:
        :param kwargs:
        :type kwargs:"""
        raise NotImplementedError



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 04-01-2021
           """

import sys

__all__ = ["get_backend_module"]

from types import ModuleType


def get_backend_module(
    project_name: str, backend_name: str = sys.platform
) -> ModuleType:
    """Returns the backend module."""
    import importlib

    modules = []
    if backend_name is not None:
        modules += [backend_name]
    elif sys.platform == "darwin":
        modules += ["darwin"]
    elif sys.platform == "win32":
        modules += ["win10"]
    else:
        modules += ["appindicator", "gtk", "xorg"]

    errors = []
    for module in modules:
        try:
            return importlib.import_module(f"{project_name}.{module}")
        except ImportError as e:
            errors.append(e)

    raise ImportError(
        f'{sys.platform} platform is not supported: {"; ".join(str(e) for e in errors)}'
    )


if __name__ == "__main__":
    print(get_backend_module("draugr"))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 23/03/2020
           """

from warg import QuadNumber

__all__ = ["xywh_to_minmax", "minmax_to_xywh"]


def xywh_to_minmax(box: QuadNumber) -> QuadNumber:
    """

    :param box:
    :type box:
    :return:
    :rtype:"""
    x1, y1, w, h = box
    return x1, y1, x1 + w, y1 + h


def minmax_to_xywh(boxes: QuadNumber) -> QuadNumber:
    """

    :param boxes:
    :type boxes:
    :return:
    :rtype:"""
    x_min, y_min, x_max, y_max = boxes
    return x_min, y_min, x_max - x_min, y_max - y_min


if __name__ == "__main__":
    quad = (2, 2, 3, 4)
    mm = xywh_to_minmax(quad)
    xywh = minmax_to_xywh(mm)
    print(quad, mm, xywh)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from typing import List

import numpy
from torch.utils.data import Dataset

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 09/10/2019
           """
__all__ = ["NonSequentialDataset"]


class NonSequentialDataset(Dataset):
    """
    * ``N`` - number of parallel environments
    * ``T`` - number of time steps explored in environments

    Dataset that flattens ``N*T*...`` arrays into ``B*...`` (where ``B`` is equal to ``N*T``) and returns
    such rows
    one by one. So basically we loose information about sequence order and we return
    for example one state, action and reward per row.

    It can be used for ``Model``'s that does not need to keep the order of events like MLP models.

    For ``LSTM`` use another implementation that will slice the dataset differently"""

    def __init__(self, *arrays: numpy.ndarray) -> None:
        """
        :param arrays: arrays to be flattened from ``N*T*...`` to ``B*...`` and returned in each call to get
        item"""
        super().__init__()
        self.arrays = [array.reshape(-1, *array.shape[2:]) for array in arrays]

    def __getitem__(self, index: int) -> List:
        return [array[index] for array in self.arrays]

    def __len__(self):
        return len(self.arrays[0])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 08-02-2021
           """

from pathlib import Path
from typing import Tuple, Union

import numpy
from scipy.io import wavfile

__all__ = ["read_normalised_wave", "write_normalised_wave"]


def read_normalised_wave(wav_file_name: Union[str, Path]) -> Tuple[int, numpy.ndarray]:
    """
    [-1..1] normalised
    """
    sampling_rate, signal = wavfile.read(str(wav_file_name))
    if signal.dtype == numpy.int16:
        num_bits = 16 - 1  # -> 16-bit wav files, -1 for sign
    elif signal.dtype == numpy.int32:
        num_bits = 32 - 1  # -> 32-bit wav files, -1 for sign
    elif signal.dtype == numpy.uint8:
        num_bits = 8
    elif signal.dtype == numpy.float32:
        return sampling_rate, signal
        # num_bits = 0
    else:
        raise NotImplementedError(f"{signal.dtype} is not supported")
    return (
        sampling_rate,
        (signal / (2**num_bits)).astype(numpy.float),
    )  # normalise by max possible val of dtype


def write_normalised_wave(
    wav_file_name: Union[str, Path],
    sampling_rate: int,
    signal: numpy.ndarray,
    dtype=numpy.float32,
) -> None:
    """
    [-1..1] normalised
    """
    assert signal.dtype == numpy.float

    if dtype == numpy.int16:
        num_bits = 16 - 1  # -> 16-bit wav files, -1 for sign
    elif dtype == numpy.int32:
        num_bits = 32 - 1  # -> 32-bit wav files, -1 for sign
    elif dtype == numpy.uint8:
        num_bits = 8
    elif dtype == numpy.float32:
        # num_bits = 0
        wavfile.write(wav_file_name, sampling_rate, signal)
        return
    else:
        raise NotImplementedError(f"{signal.dtype} is not supported")
    wavfile.write(
        str(wav_file_name), sampling_rate, signal * (2**num_bits)
    )  # unnormalise by max possible val of dtype



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 14-03-2021
           """

from typing import Tuple

import numpy

from warg import Number

__all__ = ["sphere_mask"]


def sphere_mask(
    h: Number,
    w: Number,
    d: Number,
    center: Tuple[Number, Number, Number] = None,
    radius: Number = None,
) -> numpy.ndarray:
    """

    :param d:
    :param h:
    :param w:
    :param center:
    :param radius:
    :return:"""
    if center is None:  # use the middle of the image
        center = (int(w / 2), int(h / 2), int(d / 2))
    if radius is None:  # use the smallest distance between the center and image walls
        radius = min(
            center[0], center[1], center[2], w - center[0], h - center[1], d - center[2]
        )

    y, x, z = numpy.ogrid[:h, :w, :d]
    dist_from_center = numpy.sqrt(
        (x - center[0]) ** 2 + (y - center[1]) ** 2 + (z - center[2]) ** 2
    )

    return dist_from_center <= radius


if __name__ == "__main__":
    print(sphere_mask(9, 9, 9))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import collections

import numpy

__author__ = "Christian Heider Nielsen"
__doc__ = """
Created on 27/04/2019

@author: cnheider
"""

__all__ = ["metrics"]


def metrics(array: numpy.ndarray) -> collections.OrderedDict:
    """

    :param array:
    :type array:
    :return:
    :rtype:"""
    d = collections.OrderedDict()
    d["mean"] = numpy.mean(array)
    d["std"] = numpy.std(array)
    d["min"] = numpy.amin(array)
    d["max"] = numpy.amax(array)
    return d



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 27-05-2021
           """

from enum import Enum
from pathlib import Path
from typing import List, Union

from pandas.core.generic import NDFrame
from sorcery import assigned_names

from apppath import ensure_existence
from warg import Number

__all__ = [
    "color_highlight_extreme",
    "ColorEnum",
    "NDFrameExtremeEnum",
    "AttrEnum",
    "color_negative_red",
]


class ColorEnum(Enum):
    """ """

    (
        purple,
        cyan,
        yellow,
        red,
        green,
        blue,
        magenta,
        pink,
        orange,
        black,
        white,
    ) = assigned_names()


class NDFrameExtremeEnum(Enum):
    """ """

    min = NDFrame.min
    max = NDFrame.max


class AttrEnum(Enum):
    """ """

    color = "color: {}"
    bg = "background-color: {}"


def color_highlight_extreme(
    s: Union[NDFrame, Number],
    color: Union[str, ColorEnum] = ColorEnum.yellow,
    attr_template: Union[str, AttrEnum] = AttrEnum.bg,
    extreme_func: Union[NDFrameExtremeEnum, callable] = NDFrameExtremeEnum.min,
) -> Union[List[str], str, NDFrame]:
    """
    highlight the maximum in a Series yellow.
    """
    if isinstance(color, ColorEnum):
        color = (
            color.value.lower()
        )  # TODO: LOWER IN THE CURRENT CASE no matter assigned names
    if isinstance(attr_template, AttrEnum):
        attr_template = attr_template.value

    attr = attr_template.format(color)
    if isinstance(s, NDFrame):
        mask = s == extreme_func(s)
        if s.ndim == 1:  # Series from .apply(axis=0) or axis=1
            return [attr if v else "" for v in mask]
        return pandas.DataFrame(
            numpy.where(mask, attr, ""), index=s.index, columns=s.columns
        )  # from .apply(axis=None)
    return attr


def color_negative_red(value):
    """
    Colors elements in a dateframe
    green if positive and red if
    negative. Does not color NaN
    values.
    """

    if value < 0:
        color = "red"
    elif value > 0:
        color = "green"
    else:
        color = "black"

    return f"color: {color}"


if __name__ == "__main__":
    import numpy
    import pandas
    import seaborn

    numpy.random.seed(24)
    df = pandas.DataFrame({"A": numpy.linspace(1, 10, 10)})
    df = pandas.concat(
        [df, pandas.DataFrame(numpy.random.randn(10, 4), columns=list("BCDE"))], axis=1
    )
    df.iloc[3, 3] = numpy.nan
    df.iloc[0, 2] = numpy.nan

    cm = seaborn.light_palette("green", as_cmap=True)
    df = df.style.background_gradient(cmap=cm).highlight_null(
        null_color="red"
    )  # element wise
    # df.style.bar(subset=['A', 'B'], color='#d65f5f')
    # df.style.bar(subset=['A', 'B'], align='mid', color=['#d65f5f', '#5fba7d'])
    df = df.applymap(color_highlight_extreme)  # .format(None, na_rep="-")
    df = df.apply(color_highlight_extreme, color="darkorange")
    df = df.apply(
        color_highlight_extreme,
        extreme_func=NDFrameExtremeEnum.max,
        color="green",
        axis=None,
    )

    html_ = df.render()
    # html_ = df.to_html()
    # from IPython.display import display, HTML
    # display((HTML(),))
    # print(tabulate(df, headers = 'keys', tablefmt = 'psql'))
    with open(ensure_existence(Path("exclude")) / "style_test.html", "w") as f:
        f.write(html_)

    """
Styler.applymap(func) for elementwise styles

Styler.apply(func, axis=0) for columnwise styles

Styler.apply(func, axis=1) for rowwise styles

Styler.apply(func, axis=None) for tablewise styles

And crucially the input and output shapes of func must match. If x is the input then func(x).shape == x.shape.
"""



from pathlib import Path

from warg import ContextWrapper


def make_user(
    username: str = "demo_user",
    password: str = None,
    *,
    add_home: bool = True,
    home_dir: Path = None,
    allow_existing_user: bool = True,
    get_sudo: bool = True,
) -> None:
    """ """
    import crypt
    import sh
    import getpass

    query = []

    if add_home:
        query += [f"-m", f"-d"]
        if home_dir:
            query += [str(home_dir)]
        else:
            query += [f"/home/{username}"]

    try:
        user_id = sh.id(["-u", username])
        if int(user_id):
            if not allow_existing_user:
                raise FileExistsError
            group_id = sh.id(["-g", username])
            print(f"user {username} exists with id {user_id} and {group_id}")
    except (ValueError, sh.ErrorReturnCode_1):
        pass
        with ContextWrapper(
            sh.contrib.sudo,
            construction_kwargs=dict(
                password=getpass.getpass(
                    prompt=f"[sudo] password for {getpass.getuser()}: "
                )
                if get_sudo
                else None,
                _with=True,
            ),
            enabled=get_sudo,
        ):
            try:
                sh.useradd(
                    query
                    + [
                        f"-p",
                        f"{crypt.crypt(password if password else input(f'new password for user {username}: '), '22')}",
                        f"{username}",
                    ]
                )
            except sh.ErrorReturnCode_9:
                pass


def remove_user(
    username: str = "demo_user", *, remove_home: bool = True, get_sudo: bool = True
) -> None:
    """ """
    import sh
    import getpass

    try:
        user_id = sh.id(["-u", username])
        if int(user_id):
            print(f"User {username} exists with id {user_id}")
            with ContextWrapper(
                sh.contrib.sudo,
                construction_kwargs=dict(
                    password=getpass.getpass(
                        prompt=f"[sudo] password for {getpass.getuser()}: "
                    )
                    if get_sudo
                    else None,
                    _with=True,
                ),
                enabled=get_sudo,
            ):
                sh.userdel((["-r"] if remove_home else []) + [f"{username}"])
                print(f"Removed user {username}")
    except (ValueError, sh.ErrorReturnCode_1):
        pass


def change_passwd(
    username: str = "demo_user",
    password: str = None,
) -> None:
    """

    :param username:
    :param password:
    """
    raise NotImplementedError
    pass  # ./passwd


def change_home_dir(username: str = "demo_user", new_home: str = None) -> None:
    """

    :param username:
    :param new_home:
    """
    raise NotImplementedError
    pass  # ./mkhomedir_helper username


if __name__ == "__main__":
    make_user()
    remove_user()



import cmd
from pathlib import Path

from warg import passes_kws_to, PropertySettings

__all__ = ["PlaybackShell", "ConfigShell"]


class PlaybackShell(cmd.Cmd):
    intro = "Type help or ? to list commands.\n"
    default_file_path = Path("playback.cmd")
    file = None

    # ----- record and playback -----
    def do_record(self, file: Path):
        """Save future commands to filename:  RECORD file.cmd"""
        if not file:
            file = self.default_file_path
        self.file = open(file, "w")

    def do_playback(self, file: Path):
        """Playback commands from a file:  PLAYBACK file.cmd"""
        self.close()
        if not file:
            file = self.default_file_path
        with open(file) as f:
            self.cmdqueue.extend(f.read().splitlines())

    def precmd(self, line):
        """

        :param line:
        :return:
        """
        line = line.lower()
        if self.file and "playback" not in line:
            print(line, file=self.file)
        return line

    def close(self):
        """ """
        if self.file:
            self.file.close()
            self.file = None

    def do_exit(self, arg):
        """If recording, stop, close file, close window, and exit:"""
        print("Exiting")
        self.close()
        return True

    def do_close(self, arg):
        """If recording, stop, close file, close window, and exit:"""
        self.do_exit(arg)


class ConfigShell(PlaybackShell):
    def get_names(self):  # Override!
        return dir(self)

    def onecmd(self, line):  # Override!
        try:
            return super().onecmd(line)
        except Exception as e:
            print(e)
            return False  # don't stop

    @passes_kws_to(cmd.Cmd.__init__)
    def __init__(self, name: str = "config", **kwargs):
        super().__init__(**kwargs)
        ConfigShell.prompt = f"({name}) "

    def add_property_options(self, ps: PropertySettings):
        for p in ps.__iter_keys__():
            prop = getattr(ps.__class__, p)
            getter = lambda *e: print(prop.fget(ps))
            getter.__doc__ = prop.fget.__doc__
            setter = lambda *e: prop.fset(ps, *e)
            setter.__doc__ = prop.fset.__doc__
            if prop.fdel:
                deleter = lambda *e: prop.fdel(ps)
                deleter.__doc__ = prop.fdel.__doc__
            else:
                deleter = None
            self.add_option(p, getter=getter, setter=setter, deleter=deleter)

    def add_option(self, key, *, getter, setter, deleter=None):
        self.add_func(f"get_{key}", getter)
        self.add_func(f"set_{key}", setter)
        if deleter:
            self.add_func(f"del_{key}", deleter)

    def add_func(self, key, func):
        k = f"do_{key}"
        assert k not in self.get_names()
        self.__setattr__(k, func)


if __name__ == "__main__":

    def ujsd():
        global A
        global SOME
        A = 99
        SOME = 99

        class Aconfig(PropertySettings):
            @property
            def some(self):
                """
                return SOME
                :return:
                :rtype:
                """
                global SOME
                return SOME

            @some.setter
            def some(self, i):
                """
                set SOME
                :return:
                :rtype:
                """
                global SOME
                SOME = i

            @some.deleter
            def some(self):
                """
                del SOME
                :return:
                :rtype:
                """
                global SOME
                del SOME

            @property
            def other(self):
                """
                return A
                :return:
                :rtype:
                """
                global A
                return A

            @other.setter
            def other(self, i):
                """
                set A
                :return:
                :rtype:
                """
                global A
                A = i

        ac = Aconfig()

        def set_A(e):
            global A
            A = e

        def get_A(e):
            global A
            print(A)

        cs = ConfigShell()
        cs.add_option("a", getter=get_A, setter=set_A)

        cs.add_property_options(ac)

        cs.cmdloop()

    ujsd()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from torch import nn

__author__ = "Christian Heider Nielsen"
__doc__ = ""

from draugr.torch_utilities.architectures.mlp import MLP


class MultiHeadedMLP(MLP):
    def __init__(self, *, heads_hidden_sizes=(32, 64), heads=(2, 1), **kwargs):
        super().__init__(**kwargs)

        assert len(heads_hidden_sizes) == len(heads)

        self._heads_hidden_sizes = heads_hidden_sizes
        self._heads = heads

        self.num_of_heads = len(self._heads)
        if self.num_of_heads > 0:
            for i in range(1, self.num_of_heads + 1):
                head_hidden = nn.Linear(
                    self._output_shape,
                    self._heads_hidden_sizes[i - 1],
                    bias=self._use_bias,
                )
                setattr(self, f"subhead{str(i)}_hidden", head_hidden)
                head = nn.Linear(
                    self._heads_hidden_sizes[i - 1],
                    self._heads[i - 1],
                    bias=self._use_bias,
                )
                setattr(self, f"subhead{str(i)}", head)
        else:
            raise ValueError("Number of heads must be >0")

    def forward(self, x, **kwargs):
        x = super().forward(x, **kwargs)

        output = []
        for i in range(1, self.num_of_heads + 1):
            head_hidden = getattr(self, f"subhead{str(i)}_hidden")
            x_s = head_hidden(x)
            head = getattr(self, f"subhead{str(i)}")
            sub_res = head(x_s)

            # if not isinstance(sub_res, list):
            #  sub_res = [sub_res]

            output.append(sub_res)

        return output



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 17-12-2020
           """

import numpy

__all__ = ["mean_square", "root_mean_square"]


def mean_square(x: numpy.ndarray) -> numpy.ndarray:
    """Mean Square value of signal `x`.
    input
    -----
    * x: signal vector
    output
    ------
    * Mean square of `x`."""

    return numpy.mean(numpy.abs(x) ** 2)


def root_mean_square(x: numpy.ndarray) -> numpy.ndarray:
    """Root Mean Square value of signal `x`.
    input
    -----
    * x: signal vector"""

    return numpy.sqrt(mean_square(x))  # faster than **.5



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 25/03/2020
           """

from pathlib import Path

with open(Path(__file__).parent / "README.md", "r") as this_init_file:
    __doc__ += this_init_file.read()

from .gamma import *
from .gray import *
from .noise import *
from .threshold import *
from .edge import *
from .clean import *
from .color import *



import cv2
import numpy

from draugr.dlib_utilities.dlib_utilities import (
    Dlib5faciallandmarksindices,
    Dlib68faciallandmarksindices,
    shape_to_ndarray,
)
from warg import Number

__all__ = ["align_face"]


def align_face(
    image,
    gray,
    rect,
    predictor,
    desired_left_eye=(0.35, 0.35),
    desired_face_size=None,  # (256, 256),
    padding: Number = 30,
    debug: bool = False,
):
    """

    :param image:
    :param gray:
    :param rect:
    :param predictor:
    :param desired_left_eye:
    :param desired_face_size:
    :return:
    """
    if desired_face_size is not None:
        desired_face_width, desired_face_height = desired_face_size
    else:
        desired_face_width, desired_face_height = (
            rect.width() + padding * 2,
            rect.height() + padding * 2,
        )  # BroadCastNone()

    face_shape = shape_to_ndarray(predictor(gray, rect))

    if len(face_shape) == 68:
        slicer = Dlib68faciallandmarksindices
    else:
        slicer = Dlib5faciallandmarksindices

    left_eye_pts = slicer.slice(face_shape, slicer.left_eye)
    right_eye_pts = slicer.slice(face_shape, slicer.right_eye)

    left_eye_center = left_eye_pts.mean(axis=0).astype(
        numpy.int
    )  # compute the center of mass for each eye
    right_eye_center = right_eye_pts.mean(axis=0).astype(numpy.int)

    d_y = right_eye_center[1] - left_eye_center[1]
    d_x = right_eye_center[0] - left_eye_center[0]
    angle = (
        numpy.degrees(numpy.arctan2(d_y, d_x)) - 180
    ).item()  # compute the angle between the eye centroids

    desired_right_eye_x = (
        1.0 - desired_left_eye[0]
    )  # compute the desired right eye x-coordinate based on the desired x-coordinate of the left eye

    # determine the scale of the new resulting image by taking
    # the ratio of the distance between eyes in the *current*
    # image to the ratio of distance between eyes in the
    # *desired* image
    dist = numpy.sqrt((d_x**2) + (d_y**2))
    desired_dist = desired_right_eye_x - desired_left_eye[0]
    desired_dist *= desired_face_width
    scale = (desired_dist / dist).item()

    eyes_center = (
        ((left_eye_center[0] + right_eye_center[0]) // 2).item(),
        ((left_eye_center[1] + right_eye_center[1]) // 2).item(),
    )  # compute center (x, y)-coordinates (i.e., the median point) between the two eyes in the input image

    rot_m = cv2.getRotationMatrix2D(
        eyes_center, angle, scale
    )  # grab the rotation matrix for rotating and scaling the face

    t_x = desired_face_width * 0.5  # update the translation component of the matrix
    t_y = desired_face_height * desired_left_eye[1]
    rot_m[0, 2] += t_x - eyes_center[0]
    rot_m[1, 2] += t_y - eyes_center[1]

    return cv2.warpAffine(
        image, rot_m, (desired_face_width, desired_face_height), flags=cv2.INTER_CUBIC
    )  # apply the affine transformation



from enum import Enum


class TaskLogonTypeEnum(Enum):
    """ """

    TASK_LOGON_NONE = 0
    TASK_LOGON_PASSWORD = 1
    TASK_LOGON_S4U = 2
    TASK_LOGON_INTERACTIVE_TOKEN = 3
    TASK_LOGON_GROUP = 4
    TASK_LOGON_SERVICE_ACCOUNT = 5
    TASK_LOGON_INTERACTIVE_TOKEN_OR_PASSWORD = 6



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 17-02-2021
           """

from functools import reduce
from typing import Sequence, Optional

import numpy
from matplotlib import pyplot

__all__ = [
    "plot_img_array",
    "plot_side_by_side",
]


def plot_img_array(
    img_array: numpy.ndarray, n_col: int = 3, *, labels: Optional[Sequence] = None
) -> None:
    """

    :param labels:
    :param img_array:
    :type img_array:
    :param n_col:
    :type n_col:
    :return:
    :rtype:"""
    n_row = len(img_array) // n_col
    if labels:
        assert len(img_array) == len(labels)
    f, plots = pyplot.subplots(
        n_row, n_col, sharex="all", sharey="all", figsize=(n_col * 4, n_row * 4)
    )
    if n_row > 1:
        for i in range(len(img_array)):
            plots[i // n_col, i % n_col].imshow(img_array[i])
            if labels:
                plots[i // n_col, i % n_col].set_title(labels[i])
    else:
        for i in range(len(img_array)):
            plots[i].imshow(img_array[i])
            if labels:
                plots[i].set_title(labels[i])


def plot_side_by_side(
    img_arrays: numpy.ndarray, *, labels: Optional[Sequence] = None
) -> None:
    """ """
    flatten_list = reduce(lambda x, y: x + y, zip(*img_arrays))

    plot_img_array(numpy.array(flatten_list), n_col=len(img_arrays), labels=labels)



# !/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 26-06-2021
           """

from draugr.misc_utilities import latextable

table_row_name = "tablerow"

rows = [
    ["Rocket", "Organisation", "LEO Payload (Tonnes)", "Maiden Flight"],
    ["Saturn V", "NASA", "140", "1967"],
    ["Space Shuttle", "NASA", "24.4", "1981"],
    ["Falcon 9 FT-Expended", "SpaceX", "22.8", "2017"],
    ["Ariane 5 ECA", "ESA", "21", "2002"],
]


def asijd() -> None:
    """
    :rtype: None
    """
    from tabulate import tabulate

    print("\nTabulate Latex:")
    print(tabulate(rows, headers="firstrow", tablefmt="latex"))


def adasdassijd() -> None:
    """
    :rtype: None
    """
    from texttable import Texttable

    table = Texttable()
    table.set_cols_align(["c"] * 4)
    table.set_deco(Texttable.HEADER | Texttable.VLINES)
    table.add_rows(rows)

    print("\nTexttable Table:")
    print(table.draw())

    print("\nTexttable Latex:")
    print(latextable.draw_latex(table, caption="A comparison of rocket features."))


def iasjduh() -> None:
    """
    :rtype: None
    """
    from texttable import Texttable

    table = Texttable()
    table.set_cols_align(["l", "r", "c"])
    table.set_cols_valign(["t", "m", "b"])
    table.add_rows(
        [
            ["Name", "Age", "Nickname"],
            ["Mr\nXavier\nHuon", 32, "Xav'"],
            ["Mr\nBaptiste\nClement", 1, "Baby"],
            ["Mme\nLouise\nBourgeau", 28, "Lou\n \nLoue"],
        ]
    )
    print(table.draw() + "\n")
    print(latextable.draw_latex(table, caption="An example table.") + "\n")

    table = Texttable()
    table.set_deco(Texttable.HEADER)
    table.set_cols_dtype(
        [
            "t",  # text
            "f",  # float (decimal)
            "e",  # float (exponent)
            "i",  # integer
            "a",
        ]
    )  # automatic
    table.set_cols_align(["l", "r", "r", "r", "l"])
    table.add_rows(
        [
            ["text", "float", "exp", "int", "auto"],
            ["abcd", "67", 654, 89, 128.001],
            ["efghijk", 67.5434, 0.654, 89.6, 12800000000000000000000.00023],
            ["lmn", 5e-78, 5e-78, 89.4, 0.000000000000128],
            ["opqrstu", 0.023, 5e78, 92.0, 12800000000000000000000],
        ]
    )
    print(table.draw() + "\n")
    print(
        latextable.draw_latex(
            table, caption="Another table.", label="table:another_table"
        )
        + "\n"
    )
    print(
        latextable.draw_latex(
            table,
            caption="A table with dropped columns.",
            label="table:dropped_column_table",
            drop_columns=["exp", "int"],
        )
    )


adasdassijd()



SERVICE_TEMPLATE = """
[Unit]
Description = {description}
After = network.target 
 
[Service]
Type = simple
User = {service_user}
Group = {service_group}
ExecStart = {executable} {service_entry_point_path}
Restart = {restart} 
SyslogIdentifier = {service_name}
RestartSec = 5
TimeoutStartSec = infinity
 
[Install]
WantedBy = {service_target}
"""

# .config/systemd/user/{app}.service
# sudo systemctl --user start {app}.service
# sudo systemctl --user restart {app}.service
# sudo systemctl --user enable {app}.service
# sudo systemctl --user status {app}.service
# sudo systemctl --user disable {app}.service
SERVICE_TEMPLATE_USER_SIMPLE = """
[Unit]
Description={app} Service

[Service]
Type=simple
ExecStart={app_path}

[Install]
WantedBy=default.target
"""

SERVICE_TEMPLATE_SIMPLE = """
[Unit]
Description={service_name} Service
After=multi-user.target

[Service]
Type=simple
ExecStart={executable} {service_entry_point_path}

[Install]
WantedBy=multi-user.target
"""

if __name__ == "__main__":
    print(
        SERVICE_TEMPLATE.format(
            description=1,
            python_path=2,
            script_path=3,
            service_name=4,
            executable=5,
            service_target=6,
            service_entry_point_path=7,
            service_user=8,
            service_group=9,
        )
    )
    print(
        SERVICE_TEMPLATE_SIMPLE.format(
            service_name_a=1, executable=2, service_entry_point_path=3
        )
    )
    print(SERVICE_TEMPLATE_USER_SIMPLE)



from enum import Enum

__all__ = [
    "TrainingScalars",
    "TrainingCurves",
    "TrainingTables",
    "TestingScalars",
    "TestingCurves",
    "TestingTables",
    "should_plot_y_log_scale",
    "should_smooth_series",
]

from sorcery import assigned_names


class TrainingScalars(Enum):
    """ """

    (
        training_loss,
        validation_loss,
        validation_accuracy,
        new_best_model,
    ) = assigned_names()


class TrainingCurves(Enum):
    """ """

    (
        validation_precision_recall,
        validation_receiver_operator_characteristic,
    ) = assigned_names()


class TrainingTables(Enum):
    """ """

    validation_confusion_matrix, validation_support = assigned_names()


class TestingScalars(Enum):
    """ """

    (
        test_accuracy,
        test_precision,
        test_recall,
        test_receiver_operator_characteristic_auc,
    ) = assigned_names()


class TestingCurves(Enum):
    """ """

    test_precision_recall, test_receiver_operator_characteristic = assigned_names()


class TestingTables(Enum):
    """ """

    test_confusion_matrix, test_support = assigned_names()


def should_plot_y_log_scale(tag: Enum) -> bool:
    """

    :param tag:
    :return:
    """
    if tag is TrainingScalars.training_loss:
        return True
    elif tag is TrainingScalars.validation_loss:
        return True
    return False


def should_smooth_series(tag: Enum) -> bool:
    """

    :param tag:
    :return:
    """
    if tag is TrainingScalars.training_loss:
        return True
    elif tag is TrainingScalars.validation_loss:
        return True
    return False


if __name__ == "__main__":
    print(TrainingScalars.training_loss.value)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "heider"
__doc__ = r"""

           Created on 01/02/2022
           """

__all__ = [
    "ThresholdTypeFlag",
    "WindowPropertyFlag",
    "DrawMatchesFlagEnum",
    "MouseEventFlag",
    "TermCriteriaFlag",
]

from enum import Flag

import cv2


class TermCriteriaFlag(Flag):
    count = (
        cv2.TERM_CRITERIA_COUNT
    )  # the maximum number of iterations or elements to compute
    eps = (
        cv2.TERM_CRITERIA_EPS
    )  # the desired accuracy or change in parameters at which the iterative algorithm stops
    max_iter = cv2.TERM_CRITERIA_MAX_ITER  # the maximum number of iterations to compute


class DrawMatchesFlagEnum(Flag):
    default = cv2.DRAW_MATCHES_FLAGS_DEFAULT
    # Output image matrix will be created (Mat::create), i.e. existing memory of output image may be reused. Two source image, matches and single keypoints will be drawn. For each keypoint only the center point will be drawn (without the circle around keypoint with keypoint size and orientation).

    over_outimg = cv2.DRAW_MATCHES_FLAGS_DRAW_OVER_OUTIMG
    # Output image matrix will not be created (Mat::create). Matches will be drawn on existing content of output image.

    not_draw_single_points = cv2.DRAW_MATCHES_FLAGS_NOT_DRAW_SINGLE_POINTS
    # Single keypoints will not be drawn.

    rich_keypoints = cv2.DRAW_MATCHES_FLAGS_DRAW_RICH_KEYPOINTS
    # For each keypoint the circle around keypoint with keypoint size and orientation will be drawn.


class ThresholdTypeFlag(Flag):
    binary = cv2.THRESH_BINARY  # dst(x,y)={maxval0if src(x,y)>threshotherwise
    inverse_binary = (
        cv2.THRESH_BINARY_INV
    )  # dst(x,y)={0maxvalif src(x,y)>threshotherwise
    truncate = (
        cv2.THRESH_TRUNC
    )  # dst(x,y)={thresholdsrc(x,y)if src(x,y)>threshotherwise
    to_zero = cv2.THRESH_TOZERO  # dst(x,y)={src(x,y)0if src(x,y)>threshotherwise
    inverse_to_zero = (
        cv2.THRESH_TOZERO_INV
    )  # dst(x,y)={0src(x,y)if src(x,y)>threshotherwise
    mask = cv2.THRESH_MASK
    otsu = (
        cv2.THRESH_OTSU
    )  # flag, use Otsu algorithm to choose the optimal threshold value
    triangle = (
        cv2.THRESH_TRIANGLE
    )  # flag, use Triangle algorithm to choose the optimal threshold value


class WindowPropertyFlag(Flag):
    fullscreen = (
        cv2.WND_PROP_FULLSCREEN
    )  # fullscreen property (can be WINDOW_NORMAL or WINDOW_FULLSCREEN).
    autosize = (
        cv2.WND_PROP_AUTOSIZE
    )  # autosize property (can be WINDOW_NORMAL or WINDOW_AUTOSIZE).
    keep_ratio = (
        cv2.WND_PROP_ASPECT_RATIO
    )  # window's aspect ration (can be set to WINDOW_FREERATIO or WINDOW_KEEPRATIO).
    opengl = cv2.WND_PROP_OPENGL  # opengl support.
    visible = cv2.WND_PROP_VISIBLE  # checks whether the window exists and is visible
    topmost = (
        cv2.WND_PROP_TOPMOST
    )  # property to toggle normal window being topmost or not


class MouseEventFlag(Flag):
    ctrl_down = cv2.EVENT_FLAG_CTRLKEY  # indicates that CTRL Key is pressed.
    shift_down = cv2.EVENT_FLAG_SHIFTKEY  # indicates that SHIFT Key is pressed.
    alt_down = cv2.EVENT_FLAG_ALTKEY  # indicates that ALT Key is pressed.
    left_down = cv2.EVENT_FLAG_LBUTTON  # indicates that the left mouse button is down.
    right_down = (
        cv2.EVENT_FLAG_RBUTTON
    )  # indicates that the right mouse button is down.
    middle_down = (
        cv2.EVENT_FLAG_MBUTTON
    )  # indicates that the middle mouse button is down.



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 16-09-2020
           """


def main() -> None:
    """
    :rtype: None
    """
    # from draugr import TerminalPlotWriter
    from draugr.drawers import SeriesScrollPlot
    import psutil

    # psutil.virtual_memory() # gives an object with many fields

    # dict(psutil.virtual_memory()._asdict()) # you can convert that object to a dictionary

    # psutil.virtual_memory().percent # you can have the percentage of used RAM

    # psutil.virtual_memory().available * 100 / psutil.virtual_memory().total # you can calculate percentage of available memory

    s = SeriesScrollPlot(window_length=100, reverse=False, overwrite=True)
    while True:
        s.draw(psutil.cpu_percent())

    # with TerminalPlotWriter() as w:
    #  while True:


#     w.scalar("cpu", psutil.cpu_percent())


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from abc import ABC, abstractmethod
from typing import Union

import numpy
import torch
from PIL import Image

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 09/10/2019
           """
__all__ = ["AudioWriterMixin"]


class AudioWriterMixin(ABC):
    """
    Writer subclass that provides an interface for 'writing' audio clips"""

    @abstractmethod
    def audio(
        self,
        tag: str,
        data: Union[numpy.ndarray, torch.Tensor, Image.Image],
        step,
        **kwargs
    ) -> None:
        """

        :param tag:
        :type tag:
        :param data:
        :type data:
        :param step:
        :type step:
        :param kwargs:
        :type kwargs:"""
        raise NotImplementedError



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 04-01-2021
           """

from typing import Sequence

import numpy
from matplotlib import pyplot

__all__ = ["dissected_channel_plot", "overlay_channel_plot"]

from draugr.scipy_utilities.subsampling import (
    fft_subsample,
    fir_subsample,
)


def overlay_channel_plot(
    signal: numpy.ndarray,
    title: str = "Channels",
    channel_names: Sequence[str] = None,
    sampling_rate: int = 16000,
    line_width: float = 0.2,
    max_resolution: int = 20000,
    color_func=pyplot.cm.rainbow,
) -> None:
    """ """
    n_channels = len(signal)
    sub_time, sub_signal = fft_subsample(signal, max_resolution, sampling_rate)

    pyplot.figure(1)
    pyplot.suptitle(title)
    alpha = 1 / len(sub_signal)
    if channel_names:
        labels = channel_names
    else:
        labels = [f"C{i}" for i in range(n_channels)]

    colors = iter(color_func(numpy.linspace(0, 1, n_channels)))
    for i, channel in enumerate(sub_signal):
        pyplot.plot(
            sub_time,
            channel,
            alpha=alpha,
            label=f"{labels[i]}",
            color=next(colors),
            linewidth=line_width,
        )
    pyplot.legend()


def dissected_channel_plot(
    signal: numpy.ndarray,
    *,
    title: str = "Channels",
    channel_names: Sequence[str] = None,
    sampling_rate: int = 16000,
    line_width: float = 0.2,
    col_size=4,
    max_resolution: int = 20000,
    color_func=pyplot.cm.rainbow,
) -> None:
    """ """
    n_channels = len(signal)
    f, axs = pyplot.subplots(
        n_channels, 1, sharex="all", sharey="all", figsize=(n_channels, col_size)
    )
    if not n_channels > 1:
        axs = [axs]

    sub_time, sub_signal = fft_subsample(signal, max_resolution, sampling_rate)

    colors = iter(color_func(numpy.linspace(0, 1, n_channels)))
    pyplot.subplots_adjust(wspace=0.2, hspace=0.5)
    pyplot.suptitle(title)
    for i in range(len(sub_signal)):
        axs[i].plot(sub_time, sub_signal[i], color=next(colors), linewidth=line_width)
        if channel_names:
            axs[i].set_title(f"{channel_names[i]}")
        else:
            axs[i].set_title(f"C{i}")


def orthogonal_stereo_channel_3d_plot(
    signal: numpy.ndarray,
    *,
    max_resolution: int = 20000,
    sampling_rate: int = 16000,
) -> None:
    """ """
    fig = pyplot.figure()
    ax = fig.add_subplot(111, projection="3d")

    sub_time, sub_signal = fir_subsample(signal, max_resolution, sampling_rate)

    ax.plot(sub_time, sub_signal[0])
    ax.plot(sub_time, numpy.zeros(len(sub_time)), zs=sub_signal[1])


def deinterleaved_channel_plot_file(wav_file):
    """ """
    import wave

    with wave.open(wav_file, "r") as wav_file:
        signal = numpy.fromstring(wav_file.readframes(-1), "Int16")

        length_signal = len(signal)
        num_channels = wav_file.getnchannels()

        channels = [
            [] for channel in range(num_channels)
        ]  # Split the data into channels
        for index, datum in enumerate(signal):
            channels[index % num_channels].append(datum)

        deinterleaved_length = length_signal // num_channels
        t = numpy.linspace(
            0,
            deinterleaved_length // wav_file.getframerate(),  # Get time from indices
            num=deinterleaved_length,
        )

        pyplot.figure(1)
        pyplot.title("Signal")
        alpha = 1 / num_channels
        for channel in channels:
            pyplot.plot(t, channel, alpha=alpha)


if __name__ == "__main__":

    def iushjaqdfu() -> None:
        """
        :rtype: None
        """
        sr = 1000
        max_res = sr * 4
        t = numpy.arange(sr * 4) / sr
        # noise = numpy.random.rand(sr * 2) * 0.001
        signal = numpy.sin(200 * 2 * numpy.pi * t)  # + noise
        dissected_channel_plot(signal, sampling_rate=sr, max_resolution=max_res)
        pyplot.show()
        overlay_channel_plot(signal, sampling_rate=sr, max_resolution=max_res)
        pyplot.show()
        orthogonal_stereo_channel_3d_plot(
            signal, sampling_rate=sr, max_resolution=max_res
        )
        pyplot.show()

    def decimate_stest() -> None:
        """
        :rtype: None
        """
        sr = 1000
        max_res = sr * 4
        t = numpy.arange(sr * 4) / sr
        # noise = numpy.random.rand(sr * 2) * 0.001
        signal = numpy.sin(200 * 2 * numpy.pi * t)  # + noise
        sub_time, sub_signal = fir_subsample(signal, max_res, signal)

        print(signal.shape, sub_signal.shape)
        print(signal[:10])
        print(sub_signal[:10])
        assert numpy.equal(signal, sub_signal).all()

    def iushjaqsfaddfu() -> None:
        """
        :rtype: None
        """
        sr = 1000
        max_res = sr * 4
        t = numpy.arange(sr * 4) / sr
        # noise = numpy.random.rand(sr * 2) * 0.001
        signal = numpy.sin(200 * 2 * numpy.pi * t)  # + noise
        deinterleaved_channel_plot_file(signal)
        pyplot.show()

    # iushjaqdfu()
    iushjaqsfaddfu()
    # decimate_stest()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import math

import numpy
from matplotlib.gridspec import GridSpec

from draugr.drawers.mpl_drawers.mpldrawer import MplDrawer

__author__ = "Christian Heider Nielsen"
__doc__ = """
Created on 27/04/2019

@author: cnheider
"""

__all__ = ["FastFourierTransformPlot"]

from matplotlib import pyplot

from draugr.tqdm_utilities import progress_bar

FLOAT_EPS = numpy.finfo(float).eps


class FastFourierTransformPlot(MplDrawer):
    """
    Plots last computed fft of data"""

    def __init__(
        self,
        n_fft: int = 64,
        sampling_rate=int(1.0 / 0.0005),
        title: str = "",
        figure_size=(9, 9),
        render: bool = True,
        **kwargs
    ):
        """

        :param n_fft:
        :type n_fft:
        :param sampling_rate:
        :type sampling_rate:
        :param title:
        :type title:
        :param placement:
        :type placement:
        :param fig_size:
        :type fig_size:
        :param render:
        :type render:"""
        super().__init__(render=render, figure_size=figure_size, **kwargs)

        if not render:
            return

        self.fig = pyplot.figure(figsize=figure_size)

        self.n_fft = n_fft
        self.abs_n_fft = (self.n_fft + 1) // 2
        self.sampling_rate = sampling_rate

        freq_bins = numpy.arange(self.n_fft)
        raw_array = numpy.zeros(self.n_fft)

        self.zeroes_padding = numpy.zeros((self.abs_n_fft, n_fft))

        gs = GridSpec(3, 1)
        (self.raw_ax, self.angle_ax, self.mag_ax) = [
            pyplot.subplot(gs[i]) for i in range(3)
        ]

        freqs = numpy.fft.fftfreq(self.n_fft, 1 / sampling_rate)

        (self.dft_raw_img,) = self.raw_ax.plot(freq_bins, raw_array)
        self.raw_ax.set_xlabel("Time (Sec)")
        self.raw_ax.set_ylabel("Amplitude")

        (self.dft_angle_img,) = self.angle_ax.plot(freq_bins, raw_array)
        self.angle_ax.set_xlabel("Phase [Hz]")
        self.angle_ax.set_ylabel("Angle (Radians)")
        self.angle_ax.set_ylim([-math.pi, math.pi])
        # self.angle_ax.set_xticks(freqs)

        (self.dft_mag_img,) = self.mag_ax.plot(freq_bins, raw_array)
        self.mag_ax.set_xlabel("Frequency [Hz]")
        self.mag_ax.set_xlabel("Magnitude (dB)")
        # self.mag_ax.set_xticks(freqs)

        pyplot.title(title)
        pyplot.tight_layout()

    def _draw(self, signal_sample: float, delta: float = 1 / 120) -> None:
        """

        :param signal_sample:
        :param delta: 1 / 60 for 60fps
        :return:"""
        raw_array = self.dft_raw_img.get_ydata()
        raw_array = numpy.hstack((signal_sample, raw_array[:-1]))
        self.dft_raw_img.set_ydata(raw_array)
        cur_lim = self.raw_ax.get_ylim()
        self.raw_ax.set_ylim(
            [min(cur_lim[0], signal_sample), max(cur_lim[1], signal_sample)]
        )

        f_coef = numpy.fft.fft(raw_array, n=self.n_fft)

        self.dft_angle_img.set_ydata(numpy.angle(f_coef))

        mag = 10 * numpy.log10((numpy.abs(f_coef) ** 2) + FLOAT_EPS)

        self.dft_mag_img.set_ydata(mag)
        self.mag_ax.set_ylim([min(mag), max(mag)])


if __name__ == "__main__":

    def a() -> None:
        """
        :rtype: None
        """
        duration_sec = 4
        mul = 1000
        sampling_Hz = 44
        sampling_rate = sampling_Hz * mul  # Hz
        delta = 1 / sampling_rate
        n_fft = 64
        s = FastFourierTransformPlot(n_fft=n_fft, sampling_rate=sampling_rate)
        for t in progress_bar(numpy.arange(0, duration_sec, delta)):
            ts = 2 * numpy.pi * t
            s1 = numpy.sin(ts * 1 * sampling_Hz / 2**4 * mul)
            s2 = numpy.sin(ts * 3 * sampling_Hz / 2**3 * mul + 0.33 * numpy.pi)
            s3 = numpy.sin(ts * 5 * sampling_Hz / 2**2 * mul + 0.66 * numpy.pi)
            signal = s1 + s2 + s3
            signal /= 3
            # signal += (numpy.random.random() - 0.5) * 2 * 1 / 2  # Noise
            s.draw(signal, delta=delta)

    a()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 23/07/2020
           """

__all__ = ["replicate", "BroadcastNone"]

from typing import Sequence, Union

import sorcery
from sorcery.core import FrameInfo

from warg import Number


def replicate(x: Union[Sequence, Number], times: int = 2) -> Sequence:
    """
    if not tuple

    :param times:
    :type times:
    :param x:
    :type x:
    :return:
    :rtype:"""
    if isinstance(x, Sequence):
        if len(x) == times:
            return x
    return (x,) * times


class BroadcastNone(type(None).__class__):
    @sorcery.spell
    def __new__(frame_info: FrameInfo, cls):
        num_assignments = len(frame_info.assigned_names(allow_one=True)[0])
        if num_assignments > 1:
            return (None,) * num_assignments
        return None


if __name__ == "__main__":

    def asdaa() -> None:
        """
        :rtype: None
        """
        print(replicate(2))
        print(replicate(2, 4))

        print(replicate((2, 3)))
        print(replicate((2, 3), times=4))

    def asd213aa() -> None:
        """
        :rtype: None
        """

        assd = BroadcastNone()
        ajis, sad, asd = BroadcastNone()

        print(assd)
        print(ajis, sad, asd)

    asd213aa()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 28-10-2020
           """

import torch

from warg import AlsoDecorator

__all__ = ["TorchDeterministicSession"]


class TorchDeterministicSession(AlsoDecorator):
    """
    # Disable torch jit tracing"""

    def __init__(self, no_side_effect: bool = True):
        self._no_side_effect = no_side_effect
        if no_side_effect:
            self.prev_state = torch.is_deterministic()

    def __enter__(self):
        torch.set_deterministic(True)
        return True

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self._no_side_effect:
            torch.set_deterministic(self.prev_state)
        else:
            torch.set_deterministic(False)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 05-03-2021
           """

__all__ = ["line_styles", "marker_styles"]

line_styles = (
    "-",  # solid
    "--",  # dashed
    ":",  # dotted
    "-.",  # dash-dot
    (0, (3, 5, 1, 5, 1, 5)),  # dashdotdotted
)
marker_styles = (
    "",  # Nothing
    # ",", # pixel
    "^",  # triangle
    "o",  # o
    "x",  # x
    "s",  # square
    # '*', # star
    "|",  # vline
)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 25-05-2021
           """
try:
    from .formatting import *
    from .from_dict import *
    from .latex_mean_std import *
    from .misc_utilities import *
    from .multi_index_utilities import *
    from .styling import *
except ImportError as ix:
    this_package_name = Path(__file__).parent.name
    this_package_reqs = (
        Path(__file__).parent.parent.parent
        / "requirements"
        / f"requirements_{this_package_name}.txt"
    )
    print(
        f"Make sure requirements is installed for {this_package_name}, see {this_package_reqs}"
    )  # TODO: PARSE WHAT is missing and print
    raise ix



import datetime
import getpass
import socket

try:
    import win32com.client
except ModuleNotFoundError as e:
    raise ModuleNotFoundError((e, "missing win32com.client"))

from draugr.os_utilities.windows_utilities.task_scheduler.enums import (
    TaskActionTypeEnum,
    TaskCreationEnum,
    TaskLogonTypeEnum,
    TaskTriggerEnum,
)

__all__ = ["new_user_logon_execute_task", "delete_task", "set_task_activity"]


def get_scheduler() -> win32com.client.CDispatch:
    scheduler = win32com.client.Dispatch("Schedule.Service")
    scheduler.Connect()
    return scheduler


def register_task(
    task_folder: win32com.client.CDispatch,
    task_definition: win32com.client.CDispatch,
    name: str,
    *,
    user: str = "",
    password: str = "",
    task_creation_type=TaskCreationEnum.TASK_CREATE_OR_UPDATE.value,
    task_logon_type=TaskLogonTypeEnum.TASK_LOGON_NONE.value,
) -> None:
    task_folder.RegisterTaskDefinition(
        name, task_definition, task_creation_type, user, password, task_logon_type
    )


def new_execute_action(
    task_def: win32com.client.CDispatch,
    action_path: str,
    action_arguments: str,
    *,
    action_id: str = "action0",
) -> None:
    action = task_def.Actions.Create(TaskActionTypeEnum.TASK_ACTION_EXEC.value)
    action.ID = action_id
    action.Path = action_path
    action.Arguments = action_arguments


def new_time_trigger(
    task_def: win32com.client.CDispatch,
    start_time: datetime.datetime = datetime.datetime.now()
    + datetime.timedelta(minutes=5),
    end_time: datetime.datetime = datetime.datetime.now()
    + datetime.timedelta(minutes=10),
) -> None:
    trigger = task_def.Triggers.Create(TaskTriggerEnum.TASK_TRIGGER_TIME.value)
    trigger.StartBoundary = start_time.isoformat()
    trigger.EndBoundary = end_time.isoformat()
    trigger.ExecutionTimeLimit = "PT5M"  # Five minutes


def new_logon_trigger(
    task_def: win32com.client.CDispatch,
    domain: str = socket.gethostname(),
    username: str = getpass.getuser(),
) -> None:
    trigger = task_def.Triggers.Create(TaskTriggerEnum.TASK_TRIGGER_LOGON.value)
    trigger.Id = "LogonTriggerId"
    trigger.UserId = f"{domain}\{username}"  # Must be a valid user account


def new_boot_trigger(task_def: win32com.client.CDispatch) -> None:
    trigger = task_def.Triggers.Create(TaskTriggerEnum.TASK_TRIGGER_BOOT.value)
    trigger.Id = "BootTriggerId"
    # trigger.Delay = "PT30S"  # 30 Seconds


def new_user_logon_execute_task(
    name: str,
    desc: str = "No description",
    action_path: str = "cmd.exe",
    action_arguments: str = '/c "exit"',
    *,
    task_folder: str = "\\User",
    stop_if_on_battery: bool = False,
) -> None:
    scheduler = get_scheduler()
    root_folder = scheduler.GetFolder(task_folder)
    task_def = scheduler.NewTask(0)
    task_def.RegistrationInfo.Description = desc
    task_def.RegistrationInfo.Author = getpass.getuser()
    task_def.Settings.Enabled = True
    task_def.Settings.StartWhenAvailable = True
    task_def.Settings.StopIfGoingOnBatteries = stop_if_on_battery

    new_logon_trigger(task_def)
    # new_boot_trigger(task_def)
    new_execute_action(task_def, action_path, action_arguments)
    register_task(root_folder, task_def, name)


def delete_task(task_name: str, *, task_folder: str = "\\User") -> None:
    scheduler = get_scheduler()
    task_folder = scheduler.GetFolder(task_folder)
    task_folder.DeleteTask(task_name, 0)


def set_task_activity(
    task_name: str, enable: bool, *, task_folder: str = "\\User"
) -> None:
    scheduler = get_scheduler()
    task_folder = scheduler.GetFolder(task_folder)
    task = task_folder.GetTask(task_name)
    task.Enabled = enable


if __name__ == "__main__":

    print(f"{socket.gethostname()}\{getpass.getuser()}")

    if True:
        delete_task("test_task")
        new_user_logon_execute_task("test_task")
        set_task_activity("test_task", False)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 03/08/2020
           """

from pathlib import Path

from draugr import INCLUDE_PROJECT_READMES

if INCLUDE_PROJECT_READMES:
    with open(Path(__file__).parent / "README.md", "r") as this_init_file:
        __doc__ += this_init_file.read()

# from .gtk_settings import *
from .theme_preferences import *



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from shutil import rmtree

__author__ = "Christian Heider Nielsen"
__doc__ = ""


def main(keep_alive: bool = True, use_asyncio: bool = True) -> str:
    """

      :param use_asyncio:
    :param keep_alive:
    :type keep_alive:
    :return:
    :rtype:"""
    from draugr.torch_utilities import launch_tensorboard

    # from draugr import IgnoreInterruptSignal
    # from contextlib import suppress
    from time import sleep

    from apppath import AppPath

    import argparse

    parser = argparse.ArgumentParser(description="Option for launching tensorboard")
    parser.add_argument("NAME", type=str, help="App name to open AppPath for")
    parser.add_argument(
        "--author", type=str, help="App author to open AppPath for", default=None
    )
    parser.add_argument(
        "--version", type=str, help="App version to open AppPath for", default=None
    )
    parser.add_argument(
        "--clean",
        action="store_true",
        default=False,
        help="Adding --clean argument will wipe tensorboard logs",
    )
    parser.add_argument(
        "--port", default=6006, help="Which port should tensorboard be served on"
    )
    args = parser.parse_args()

    project_app_path = AppPath(args.NAME, args.author, args.version)
    log_dir = project_app_path.user_log

    if args.clean:
        print(f"Wiping {log_dir}")
        if log_dir.exists():
            rmtree(str(log_dir))
        else:
            log_dir.mkdir()

    address = launch_tensorboard(log_dir, args.port)

    if keep_alive:
        print(f"tensorboard address: {address} for log_dir {log_dir}")

        if use_asyncio:
            import asyncio

            async def work():
                """ """
                while True:
                    await asyncio.sleep(1)
                    # print("Task Executed")

            loop = asyncio.get_event_loop()
            try:
                asyncio.ensure_future(work())
                loop.run_forever()
            except KeyboardInterrupt:
                pass
            finally:
                print("Closing Loop")
                loop.close()
        else:
            # with IgnoreInterruptSignal(): # Do not block
            while True:
                sleep(10)

    return address


if __name__ == "__main__":
    import webbrowser

    """
Type Name

Class Name

Notes

'mozilla'

Mozilla('mozilla')

'firefox'

Mozilla('mozilla')

'netscape'

Mozilla('netscape')

'galeon'

Galeon('galeon')

'epiphany'

Galeon('epiphany')

'skipstone'

BackgroundBrowser('skipstone')

'kfmclient'

Konqueror()

(1)

'konqueror'

Konqueror()

(1)

'kfm'

Konqueror()

(1)

'mosaic'

BackgroundBrowser('mosaic')

'opera'

Opera()

'grail'

Grail()

'links'

GenericBrowser('links')

'elinks'

Elinks('elinks')

'lynx'

GenericBrowser('lynx')

'w3m'

GenericBrowser('w3m')

'windows-default'

WindowsDefault

(2)

'macosx'

MacOSX('default')

(3)

'safari'

MacOSX('safari')

(3)

'google-chrome'

Chrome('google-chrome')

'chrome'

Chrome('chrome')

'chromium'

Chromium('chromium')

'chromium-browser'

Chromium('chromium-browser')



"""

    # Open URL in a new tab, if a browser window is already open.
    webbrowser.open_new_tab(main())

    ## Open URL in new window, raising the window if possible.
    # webbrowser.open_new(url)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

          If a tensorboard plugin exist when you are reading this, go use that :)
          
          TODO: Finish

           Created on 31-03-2021
           """

if __name__ == "__main__":

    def alt():
        """

        :return:
        """
        import tensorflow as tf
        import numpy

        import textwrap
        import re
        import io
        import itertools
        import matplotlib

        class SaverHook(tf.train.SessionRunHook):
            """
            Saves a confusion matrix as a Summary so that it can be shown in tensorboard
            """

            def __init__(self, labels, confusion_matrix_tensor_name, summary_writer):
                """Initializes a `SaveConfusionMatrixHook`.

                :param labels: Iterable of String containing the labels to print for each
                               row/column in the confusion matrix.
                :param confusion_matrix_tensor_name: The name of the tensor containing the confusion
                                                     matrix
                :param summary_writer: The summary writer that will save the summary
                """
                self.confusion_matrix_tensor_name = confusion_matrix_tensor_name
                self.labels = labels
                self._summary_writer = summary_writer

            def end(self, session):
                """

                :param session:
                """
                cm = (
                    tf.get_default_graph()
                    .get_tensor_by_name(self.confusion_matrix_tensor_name + ":0")
                    .eval(session=session)
                    .astype(int)
                )
                global_step = tf.train.get_global_step().eval(session=session)
                figure = self._plot_confusion_matrix(cm)
                summary = self._figure_to_summary(figure)
                self._summary_writer.add_summary(summary, global_step)

            def _figure_to_summary(self, fig):
                """
                Converts a matplotlib figure ``fig`` into a TensorFlow Summary object
                that can be directly fed into ``Summary.FileWriter``.
                :param fig: A ``matplotlib.figure.Figure`` object.
                :return: A TensorFlow ``Summary`` protobuf object containing the plot image
                         as a image summary.
                """

                # attach a new canvas if not exists
                if fig.canvas is None:
                    matplotlib.backends.backend_agg.FigureCanvasAgg(fig)

                fig.canvas.draw()
                w, h = fig.canvas.get_width_height()

                # get PNG data from the figure
                png_buffer = io.BytesIO()
                fig.canvas.print_png(png_buffer)
                png_encoded = png_buffer.getvalue()
                png_buffer.close()

                summary_image = tf.Summary.Image(
                    height=h,
                    width=w,
                    colorspace=4,  # RGB-A
                    encoded_image_string=png_encoded,
                )
                summary = tf.Summary(
                    value=[
                        tf.Summary.Value(
                            tag=self.confusion_matrix_tensor_name, image=summary_image
                        )
                    ]
                )
                return summary

            def _plot_confusion_matrix(self, cm):
                """
                    :param cm: A confusion matrix: A square ```numpy array``` of the same size as self.labels
                `   :return:  A ``matplotlib.figure.Figure`` object with a numerical and graphical representation of the cm array
                """
                numClasses = len(self.labels)

                fig = matplotlib.figure.Figure(
                    figsize=(numClasses, numClasses),
                    dpi=100,
                    facecolor="w",
                    edgecolor="k",
                )
                ax = fig.add_subplot(1, 1, 1)
                ax.imshow(cm, cmap="Oranges")

                classes = [
                    re.sub(r"([a-z](?=[A-Z])|[A-Z](?=[A-Z][a-z]))", r"\1 ", x)
                    for x in self.labels
                ]
                classes = ["\n".join(textwrap.wrap(l, 20)) for l in classes]

                tick_marks = numpy.arange(len(classes))

                ax.set_xlabel("Predicted")
                ax.set_xticks(tick_marks)
                ax.set_xticklabels(classes, rotation=-90, ha="center")
                ax.xaxis.set_label_position("bottom")
                ax.xaxis.tick_bottom()

                ax.set_ylabel("True Label")
                ax.set_yticks(tick_marks)
                ax.set_yticklabels(classes, va="center")
                ax.yaxis.set_label_position("left")
                ax.yaxis.tick_left()

                for i, j in itertools.product(range(numClasses), range(numClasses)):
                    ax.text(
                        j,
                        i,
                        int(cm[i, j]) if cm[i, j] != 0 else ".",
                        horizontalalignment="center",
                        verticalalignment="center",
                        color="black",
                    )
                fig.set_tight_layout(True)
                return fig

    def asda():
        """

        :return:
        """
        import io
        import itertools
        from packaging import version

        import tensorflow as tf

        from matplotlib import pyplot
        import numpy

        test_pred_raw = []
        test_labels = []
        class_names = []
        train_labels = []
        train_images = []

        def plot_to_image(figure):
            """Converts the matplotlib plot specified by 'figure' to a PNG image and
            returns it. The supplied figure is closed and inaccessible after this call."""
            # Save the plot to a PNG in memory.
            buf = io.BytesIO()
            pyplot.savefig(buf, format="png")
            # Closing the figure prevents it from being displayed directly inside
            # the notebook.
            pyplot.close(figure)
            buf.seek(0)
            # Convert PNG buffer to TF image
            image = tf.image.decode_png(buf.getvalue(), channels=4)
            # Add the batch dimension
            image = tf.expand_dims(image, 0)
            return image

        def image_grid():
            """Return a 5x5 grid of the MNIST images as a matplotlib figure."""
            # Create a figure to contain the plot.
            figure = pyplot.figure(figsize=(10, 10))
            for i in range(25):
                # Start next subplot.
                pyplot.subplot(5, 5, i + 1, title=class_names[train_labels[i]])
                pyplot.xticks([])
                pyplot.yticks([])
                pyplot.grid(False)
                pyplot.imshow(train_images[i], cmap=pyplot.cm.binary)

            return figure

        def plot_confusion_matrix(cm, class_names):
            """
            Returns a matplotlib figure containing the plotted confusion matrix.

            Args:
              cm (array, shape = [n, n]): a confusion matrix of integer classes
              class_names (array, shape = [n]): String names of the integer classes
            """
            figure = pyplot.figure(figsize=(8, 8))
            pyplot.imshow(cm, interpolation="nearest", cmap=pyplot.cm.Blues)
            pyplot.title("Confusion matrix")
            pyplot.colorbar()
            tick_marks = numpy.arange(len(class_names))
            pyplot.xticks(tick_marks, class_names, rotation=45)
            pyplot.yticks(tick_marks, class_names)

            # Compute the labels from the normalized confusion matrix.
            labels = numpy.around(
                cm.astype("float") / cm.sum(axis=1)[:, numpy.newaxis], decimals=2
            )

            # Use white text if squares are dark; otherwise black.
            threshold = cm.max() / 2.0
            for i, j in itertools.product(range(cm.shape[0]), range(cm.shape[1])):
                color = "white" if cm[i, j] > threshold else "black"
                pyplot.text(
                    j, i, labels[i, j], horizontalalignment="center", color=color
                )

            pyplot.tight_layout()
            pyplot.ylabel("True label")
            pyplot.xlabel("Predicted label")
            return figure

        """


test_pred = numpy.argmax(test_pred_raw, axis=1)

# Calculate the confusion matrix.
cm = sklearn.metrics.confusion_matrix(test_labels, test_pred)
# Log the confusion matrix as an image summary.
figure = plot_confusion_matrix(cm, class_names=class_names)
cm_image = plot_to_image(figure)

# Log the confusion matrix as an image summary.
with file_writer_cm.as_default():
tf.summary.image("Confusion Matrix", cm_image, step=epoch)
"""



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

          A few low dimensional orders common in raster-grids

           Created on 28-03-2021
           """

import torch

__all__ = [
    "hwc_to_chw_tensor",
    "chw_to_hwc_tensor",
    "nhwc_to_nchw_tensor",
    "nchw_to_nhwc_tensor",
    "nthwc_to_ntchw_tensor",
    "ntchw_to_nthwc_tensor",
]


def hwc_to_chw_tensor(tensor: torch.Tensor) -> torch.Tensor:
    """

    :param tensor:
    :return:
    :rtype:"""
    assert len(tensor.shape) == 3
    return tensor.permute(2, 0, 1)


def chw_to_hwc_tensor(tensor: torch.Tensor) -> torch.Tensor:
    """

    :param tensor:
    :return:
    :rtype:"""
    assert len(tensor.shape) == 3
    return tensor.permute(1, 2, 0)


def nhwc_to_nchw_tensor(tensor: torch.Tensor) -> torch.Tensor:
    """

    :param tensor:
    :return:
    :rtype:"""
    assert len(tensor.shape) == 4
    return tensor.permute(0, 3, 1, 2)


def nchw_to_nhwc_tensor(tensor: torch.Tensor) -> torch.Tensor:
    """

    :param tensor:
    :return:
    :rtype:"""
    assert len(tensor.shape) == 4
    return tensor.permute(0, 2, 3, 1)


def nthwc_to_ntchw_tensor(tensor: torch.Tensor) -> torch.Tensor:
    """

    :param tensor:
    :return:
    :rtype:"""
    assert len(tensor.shape) == 5
    return tensor.permute(0, 1, 4, 2, 3)


def ntchw_to_nthwc_tensor(tensor: torch.Tensor) -> torch.Tensor:
    """

    :param tensor:
    :return:
    :rtype:"""
    assert len(tensor.shape) == 5
    return tensor.permute(0, 1, 3, 4, 2)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 02-12-2020
           """

from contextlib import contextmanager
from itertools import tee

from torch.nn import Module

from draugr.torch_utilities.optimisation.parameters.freezing.parameters import (
    freeze_parameters,
)

__all__ = ["freeze_model", "frozen_model"]


def freeze_model(model: Module, value: bool = None, recurse: bool = True) -> None:
    """

    :param model:
    :type model:
    :param recurse:
    :param value:
    :return:"""
    freeze_parameters(model.parameters(recurse), value)


@contextmanager
def frozen_model(model: Module, recurse: bool = True, enabled: bool = True) -> None:
    """

    :param enabled:
    :type enabled:
    :param model:
    :param recurse:
    :return:"""
    params_1, params_2 = tee(model.parameters(recurse))
    if enabled:
        freeze_parameters(params_1, True)
    yield True
    if enabled:
        freeze_parameters(params_2, False)


if __name__ == "__main__":
    from torch import nn

    def asda() -> None:
        """
        :rtype: None
        """
        a = nn.Linear(10, 5)
        print(a.weight.requires_grad)
        with frozen_model(a):
            print(a.weight.requires_grad)
        print(a.weight.requires_grad)

    asda()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


__author__ = "heider"
__doc__ = r"""

           Created on 01/02/2022
           """

from .color_conversion import *
from .css_colors import *
from .label_colors import *



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 17-05-2021
           """

from .all import *
from .cpu import *
from .ram import *
from .gpu import *



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 25/03/2020
           """

import sys
from typing import Sequence, TextIO, Tuple, Union

import numpy
import torch
from torch import nn

from draugr.torch_utilities.optimisation.parameters.counting import get_num_parameters

__all__ = ["get_model_complexity_info", "MODULES_MAPPING"]


def get_model_complexity_info(
    model: torch.nn.Module,
    input_res: Tuple,
    print_per_layer_stat: bool = True,
    as_strings: bool = True,
    input_constructor: callable = None,
    ost: TextIO = sys.stdout,
) -> Union[Tuple[int, int], Tuple[str, str]]:
    """

    :param model:
    :type model:
    :param input_res:
    :type input_res:
    :param print_per_layer_stat:
    :type print_per_layer_stat:
    :param as_strings:
    :type as_strings:
    :param input_constructor:
    :type input_constructor:
    :param ost:
    :type ost:
    :return:
    :rtype:"""
    assert isinstance(input_res, Sequence)
    assert len(input_res) >= 1
    assert isinstance(model, nn.Module)

    flops_model = add_flops_counting_methods(model)
    flops_model.eval().start_flops_count()
    if input_constructor:
        _ = flops_model(**input_constructor(input_res))
    else:
        try:
            batch = torch.ones(()).new_empty(
                (1, *input_res),
                dtype=next(flops_model.parameters()).dtype,
                device=next(flops_model.parameters()).device,
            )
        except StopIteration:
            batch = torch.ones(()).new_empty((1, *input_res))

        _ = flops_model(batch)

    if print_per_layer_stat:
        print_model_with_flops(flops_model, ost=ost)

    flops_count = flops_model.compute_average_flops_cost()
    params_count = get_num_parameters(flops_model, only_trainable=True)
    flops_model.stop_flops_count()

    if as_strings:
        return flops_to_string(flops_count), params_to_string(params_count)

    return flops_count, params_count


def flops_to_string(flops: int, units: str = "GMac", precision: int = 2) -> str:
    """

    :param flops:
    :type flops:
    :param units:
    :type units:
    :param precision:
    :type precision:
    :return:
    :rtype:"""
    if units is None:
        if flops // 10**9 > 0:
            return f"{str(round(flops / 10. ** 9, precision))} GMac"
        elif flops // 10**6 > 0:
            return f"{str(round(flops / 10. ** 6, precision))} MMac"
        elif flops // 10**3 > 0:
            return f"{str(round(flops / 10. ** 3, precision))} KMac"
        else:
            return f"{str(flops)} Mac"
    else:
        if units == "GMac":
            return f"{str(round(flops / 10. ** 9, precision))} {units}"
        elif units == "MMac":
            return f"{str(round(flops / 10. ** 6, precision))} {units}"
        elif units == "KMac":
            return f"{str(round(flops / 10. ** 3, precision))} {units}"
        else:
            return f"{str(flops)} Mac"


def params_to_string(params_num) -> str:
    """

    :param params_num:
    :type params_num:
    :return:
    :rtype:"""
    if params_num // 10**6 > 0:
        return str(round(params_num / 10**6, 2)) + " M"
    elif params_num // 10**3:
        return str(round(params_num / 10**3, 2)) + " k"
    else:
        return str(params_num)


def print_model_with_flops(
    model, units: str = "GMac", precision: int = 3, ost: TextIO = sys.stdout
) -> None:
    """

    :param model:
    :type model:
    :param units:
    :type units:
    :param precision:
    :type precision:
    :param ost:
    :type ost:
    :return:
    :rtype:"""
    total_flops = model.compute_average_flops_cost()

    def accumulate_flops(self) -> int:
        """

        :param self:
        :type self:
        :return:
        :rtype:"""
        if is_supported_instance(self):
            return self.__flops__ / model.__batch_counter__
        else:
            sum = 0
            for m in self.children():
                sum += m.accumulate_flops()
            return sum

    def flops_repr(self) -> str:
        """

        :param self:
        :type self:
        :return:
        :rtype:"""
        accumulated_flops_cost = self.accumulate_flops()
        return ", ".join(
            [
                flops_to_string(
                    accumulated_flops_cost, units=units, precision=precision
                ),
                f"{accumulated_flops_cost / total_flops:.3%} MACs",
                self.original_extra_repr(),
            ]
        )

    def add_extra_repr(m) -> None:
        """

        :param m:
        :type m:"""
        m.accumulate_flops = accumulate_flops.__get__(m)
        flops_extra_repr = flops_repr.__get__(m)
        if m.extra_repr != flops_extra_repr:
            m.original_extra_repr = m.extra_repr
            m.extra_repr = flops_extra_repr
            assert m.extra_repr != m.original_extra_repr

    def del_extra_repr(m) -> None:
        """

        :param m:
        :type m:"""
        if hasattr(m, "original_extra_repr"):
            m.extra_repr = m.original_extra_repr
            del m.original_extra_repr
        if hasattr(m, "accumulate_flops"):
            del m.accumulate_flops

    model.apply(add_extra_repr)
    print(model, file=ost)
    model.apply(del_extra_repr)


def add_flops_counting_methods(net_main_module: torch.nn.Module) -> torch.nn.Module:
    """

    :param net_main_module:
    :type net_main_module:
    :return:
    :rtype:"""
    # adding additional methods to the existing module object,
    # this is done this way so that each function has access to self object
    net_main_module.start_flops_count = start_flops_count.__get__(net_main_module)
    net_main_module.stop_flops_count = stop_flops_count.__get__(net_main_module)
    net_main_module.reset_flops_count = reset_flops_count.__get__(net_main_module)
    net_main_module.compute_average_flops_cost = compute_average_flops_cost.__get__(
        net_main_module
    )

    net_main_module.reset_flops_count()

    # Adding variables necessary for masked flops computation
    net_main_module.apply(add_flops_mask_variable_or_reset)

    return net_main_module


def compute_average_flops_cost(self) -> float:
    """
    A method that will be available after add_flops_counting_methods() is called
    on a desired net object.

    Returns current mean flops consumption per image."""

    batches_count = self.__batch_counter__
    flops_sum = 0
    for module in self.modules():
        if is_supported_instance(module):
            flops_sum += module.__flops__

    return flops_sum / batches_count


def start_flops_count(self) -> None:
    """
    A method that will be available after add_flops_counting_methods() is called
    on a desired net object.

    Activates the computation of mean flops consumption per image.
    Call it before you run the network."""
    add_batch_counter_hook_function(self)
    self.apply(add_flops_counter_hook_function)


def stop_flops_count(self) -> None:
    """
    A method that will be available after add_flops_counting_methods() is called
    on a desired net object.

    Stops computing the mean flops consumption per image.
    Call whenever you want to pause the computation."""
    remove_batch_counter_hook_function(self)
    self.apply(remove_flops_counter_hook_function)


def reset_flops_count(self) -> None:
    """
    A method that will be available after add_flops_counting_methods() is called
    on a desired net object.

    Resets statistics computed so far."""
    add_batch_counter_variables_or_reset(self)
    self.apply(add_flops_counter_variable_or_reset)


def add_flops_mask(module: torch.nn.Module, mask) -> None:
    """

    :param module:
    :type module:
    :param mask:
    :type mask:"""

    def add_flops_mask_func(module: torch.nn.Module) -> None:
        """

        :param module:
        :type module:"""
        if isinstance(module, torch.nn.Conv2d):
            module.__mask__ = mask

    module.apply(add_flops_mask_func)


def remove_flops_mask(module: torch.nn.Module) -> None:
    """

    :param module:
    :type module:"""
    module.apply(add_flops_mask_variable_or_reset)


# ---- Internal functions
def empty_flops_counter_hook(module: torch.nn.Module, input, output) -> None:
    """

    :param module:
    :type module:
    :param input:
    :type input:
    :param output:
    :type output:"""
    module.__flops__ += 0


def upsample_flops_counter_hook(module: torch.nn.Module, input, output) -> None:
    """

    :param module:
    :type module:
    :param input:
    :type input:
    :param output:
    :type output:"""
    output_size = output[0]
    batch_size = output_size.shape[0]
    output_elements_count = batch_size
    for val in output_size.shape[1:]:
        output_elements_count *= val
    module.__flops__ += int(output_elements_count)


def relu_flops_counter_hook(module: torch.nn.Module, input, output) -> None:
    """

    :param module:
    :type module:
    :param input:
    :type input:
    :param output:
    :type output:"""
    active_elements_count = output.numel()
    module.__flops__ += int(active_elements_count)


def linear_flops_counter_hook(module: torch.nn.Module, input, output) -> None:
    """

    :param module:
    :type module:
    :param input:
    :type input:
    :param output:
    :type output:"""
    input = input[0]
    batch_size = input.shape[0]
    module.__flops__ += int(batch_size * input.shape[1] * output.shape[1])


def pool_flops_counter_hook(module: torch.nn.Module, input, output) -> None:
    """

    :param module:
    :type module:
    :param input:
    :type input:
    :param output:
    :type output:"""
    input = input[0]
    module.__flops__ += int(numpy.prod(input.shape))


def bn_flops_counter_hook(module: torch.nn.Module, input, output) -> None:
    """

    :param module:
    :type module:
    :param input:
    :type input:
    :param output:
    :type output:"""
    # module.affine
    input = input[0]

    batch_flops = numpy.prod(input.shape)
    if module.affine:
        batch_flops *= 2
    module.__flops__ += int(batch_flops)


def deconv_flops_counter_hook(conv_module: torch.nn.Module, input, output) -> None:
    """

    :param conv_module:
    :type conv_module:
    :param input:
    :type input:
    :param output:
    :type output:"""
    # Can have multiple inputs, getting the first one
    input = input[0]

    batch_size = input.shape[0]
    input_height, input_width = input.shape[2:]

    kernel_height, kernel_width = conv_module.kernel_size
    in_channels = conv_module.in_channels
    out_channels = conv_module.out_channels
    groups = conv_module.groups

    filters_per_channel = out_channels // groups
    conv_per_position_flops = (
        kernel_height * kernel_width * in_channels * filters_per_channel
    )

    active_elements_count = batch_size * input_height * input_width
    overall_conv_flops = conv_per_position_flops * active_elements_count
    bias_flops = 0
    if conv_module.bias is not None:
        output_height, output_width = output.shape[2:]
        bias_flops = out_channels * batch_size * output_height * output_height
    overall_flops = overall_conv_flops + bias_flops

    conv_module.__flops__ += int(overall_flops)


def conv_flops_counter_hook(conv_module: torch.nn.Module, input, output) -> None:
    """

    :param conv_module:
    :type conv_module:
    :param input:
    :type input:
    :param output:
    :type output:"""
    # Can have multiple inputs, getting the first one
    input = input[0]

    batch_size = input.shape[0]
    output_dims = list(output.shape[2:])

    kernel_dims = list(conv_module.kernel_size)
    in_channels = conv_module.in_channels
    out_channels = conv_module.out_channels
    groups = conv_module.groups

    filters_per_channel = out_channels // groups
    conv_per_position_flops = (
        numpy.prod(kernel_dims) * in_channels * filters_per_channel
    )

    active_elements_count = batch_size * numpy.prod(output_dims)

    if conv_module.__mask__ is not None:
        # (b, 1, h, w)
        output_height, output_width = output_dims
        flops_mask = conv_module.__mask__.expand(
            batch_size, 1, output_height, output_width
        )
        active_elements_count = flops_mask.sum()

    overall_conv_flops = conv_per_position_flops * active_elements_count

    bias_flops = 0

    if conv_module.bias is not None:
        bias_flops = out_channels * active_elements_count

    overall_flops = overall_conv_flops + bias_flops

    conv_module.__flops__ += int(overall_flops)


def batch_counter_hook(module: torch.nn.Module, input, output) -> None:
    """

    :param module:
    :type module:
    :param input:
    :type input:
    :param output:
    :type output:"""
    batch_size = 1
    if len(input) > 0:
        # Can have multiple inputs, getting the first one
        input = input[0]
        batch_size = len(input)
    else:
        print(
            "Warning! No positional inputs found for a module, assuming batch size is 1."
        )
    module.__batch_counter__ += batch_size


def add_batch_counter_variables_or_reset(module: torch.nn.Module) -> None:
    """

    :param module:
    :type module:"""
    module.__batch_counter__ = 0


def add_batch_counter_hook_function(module: torch.nn.Module) -> None:
    """

    :param module:
    :type module:
    :return:
    :rtype:"""
    if hasattr(module, "__batch_counter_handle__"):
        return

    handle = module.register_forward_hook(batch_counter_hook)
    module.__batch_counter_handle__ = handle


def remove_batch_counter_hook_function(module: torch.nn.Module) -> None:
    """

    :param module:
    :type module:"""
    if hasattr(module, "__batch_counter_handle__"):
        module.__batch_counter_handle__.remove()
        del module.__batch_counter_handle__


def add_flops_counter_variable_or_reset(module: torch.nn.Module) -> None:
    """

    :param module:
    :type module:"""
    if is_supported_instance(module):
        module.__flops__ = 0


def is_supported_instance(module: torch.nn.Module) -> bool:
    """

    :param module:
    :type module:
    :return:
    :rtype:"""
    if type(module) in MODULES_MAPPING:
        return True
    return False


def add_flops_counter_hook_function(module: torch.nn.Module) -> None:
    """

    :param module:
    :type module:
    :return:
    :rtype:"""
    if is_supported_instance(module):
        if hasattr(module, "__flops_handle__"):
            return
        handle = module.register_forward_hook(MODULES_MAPPING[type(module)])
        module.__flops_handle__ = handle


def remove_flops_counter_hook_function(module) -> None:
    """

    :param module:
    :type module:"""
    if is_supported_instance(module):
        if hasattr(module, "__flops_handle__"):
            module.__flops_handle__.remove()
            del module.__flops_handle__


# --- Masked flops counting

# Also being run in the initialization
def add_flops_mask_variable_or_reset(module) -> None:
    """

    :param module:
    :type module:"""
    if is_supported_instance(module):
        module.__mask__ = None


MODULES_MAPPING = {
    # convolutions
    torch.nn.Conv1d: conv_flops_counter_hook,
    torch.nn.Conv2d: conv_flops_counter_hook,
    torch.nn.Conv3d: conv_flops_counter_hook,
    # activations
    torch.nn.ReLU: relu_flops_counter_hook,
    torch.nn.PReLU: relu_flops_counter_hook,
    torch.nn.ELU: relu_flops_counter_hook,
    torch.nn.LeakyReLU: relu_flops_counter_hook,
    torch.nn.ReLU6: relu_flops_counter_hook,
    # poolings
    torch.nn.MaxPool1d: pool_flops_counter_hook,
    torch.nn.AvgPool1d: pool_flops_counter_hook,
    torch.nn.AvgPool2d: pool_flops_counter_hook,
    torch.nn.MaxPool2d: pool_flops_counter_hook,
    torch.nn.MaxPool3d: pool_flops_counter_hook,
    torch.nn.AvgPool3d: pool_flops_counter_hook,
    nn.AdaptiveMaxPool1d: pool_flops_counter_hook,
    nn.AdaptiveAvgPool1d: pool_flops_counter_hook,
    nn.AdaptiveMaxPool2d: pool_flops_counter_hook,
    nn.AdaptiveAvgPool2d: pool_flops_counter_hook,
    nn.AdaptiveMaxPool3d: pool_flops_counter_hook,
    nn.AdaptiveAvgPool3d: pool_flops_counter_hook,
    # BNs
    torch.nn.BatchNorm1d: bn_flops_counter_hook,
    torch.nn.BatchNorm2d: bn_flops_counter_hook,
    torch.nn.BatchNorm3d: bn_flops_counter_hook,
    # FC
    torch.nn.Linear: linear_flops_counter_hook,
    # Upscale
    torch.nn.Upsample: upsample_flops_counter_hook,
    # Deconvolution
    torch.nn.ConvTranspose2d: deconv_flops_counter_hook,
}

if __name__ == "__main__":
    import torchvision.models as models
    import torch

    with torch.cuda.device(0):
        net = models.densenet161()
        macs, params = get_model_complexity_info(
            net, (3, 224, 224), as_strings=True, print_per_layer_stat=True
        )
        print(f'{"Computational complexity: ":<30}  {macs:<8}')
        print(f'{"Number of parameters: ":<30}  {params:<8}')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import csv
from typing import Any, TextIO, Tuple

from apppath import ensure_existence
from draugr import PROJECT_APP_PATH
from draugr.writers.writer import Writer

__author__ = "Christian Heider Nielsen"
__doc__ = """
Created on 27/04/2019

@author: cnheider
"""

__all__ = ["CSVWriter"]

from pathlib import Path


class CSVWriter(Writer):
    """ """

    @staticmethod
    def get_csv_writer(path: Path = Path.home() / "Models") -> Tuple[TextIO, Any]:
        """

        :param path:
        :type path:
        :return:
        :rtype:"""
        if path.is_dir() or path.suffix != ".csv":
            path /= "log.csv"
        csv_file = open(
            str(
                ensure_existence(path, overwrite_on_wrong_type=True, declare_file=True)
            ),
            mode="a",
        )
        return csv_file, csv.writer(csv_file)

    def _scalar(self, tag: str, value: float, step: int) -> None:
        self._write(step, tag, value)

    def __init__(self, path, **kwargs):
        super().__init__(**kwargs)
        self._path = path
        self._file = None
        self._writer = None

    def _open(self):
        self._file, self._writer = self.get_csv_writer(self._path)
        return self

    def _close(self, exc_type=None, exc_val=None, exc_tb=None) -> None:
        self._file.close()

    def _write(self, *d) -> None:
        self._writer.writerow(d)


if __name__ == "__main__":
    with CSVWriter(PROJECT_APP_PATH.user_log / "test") as p:
        p.scalar("s", 2)



import math

import cv2
import numpy

from draugr.opencv_utilities import LineTypeEnum


# ============================================================================


def ellipse_bbox(h, k, a, b, theta):
    ux = a * math.cos(theta)
    uy = a * math.sin(theta)
    vx = b * math.cos(theta + math.pi / 2)
    vy = b * math.sin(theta + math.pi / 2)
    box_halfwidth = numpy.ceil(math.sqrt(ux**2 + vx**2))
    box_halfheight = numpy.ceil(math.sqrt(uy**2 + vy**2))
    return (
        (int(h - box_halfwidth), int(k - box_halfheight)),
        (int(h + box_halfwidth), int(k + box_halfheight)),
    )


# ----------------------------------------------------------------------------

# Rotated elliptical gradient - slow, Python-only approach
def make_gradient_v1(width, height, h, k, a, b, theta):
    # Precalculate constants
    st, ct = math.sin(theta), math.cos(theta)
    aa, bb = a**2, b**2

    weights = numpy.zeros((height, width), numpy.float64)
    for y in range(height):
        for x in range(width):
            weights[y, x] = (((x - h) * ct + (y - k) * st) ** 2) / aa + (
                ((x - h) * st - (y - k) * ct) ** 2
            ) / bb

    return numpy.clip(1.0 - weights, 0, 1)


# ----------------------------------------------------------------------------

# Rotated elliptical gradient - faster, vectorized numpy approach
def make_gradient_v2(width, height, h, k, a, b, theta):
    # Precalculate constants
    st, ct = math.sin(theta), math.cos(theta)
    aa, bb = a**2, b**2

    # Generate (x,y) coordinate arrays
    y, x = numpy.mgrid[-k : height - k, -h : width - h]
    # Calculate the weight for each pixel
    weights = (((x * ct + y * st) ** 2) / aa) + (((x * st - y * ct) ** 2) / bb)

    return numpy.clip(1.0 - weights, 0, 1)


# ============================================================================
if __name__ == "__main__":

    from pathlib import Path
    from apppath import ensure_existence
    from draugr.opencv_utilities import show_image

    basep = ensure_existence(Path("exclude"))

    def draw_image(a, b, theta, inner_scale, save_intermediate=False):
        # Calculate the image size needed to draw this and center the ellipse
        _, (h, k) = ellipse_bbox(0, 0, a, b, theta)  # Ellipse center
        h += 2  # Add small margin
        k += 2  # Add small margin
        width, height = (h * 2 + 1, k * 2 + 1)  # Canvas size

        # Parameters defining the two ellipses for OpenCV (a RotatedRect structure)
        ellipse_outer = ((h, k), (a * 2, b * 2), math.degrees(theta))
        ellipse_inner = (
            (h, k),
            (a * 2 * inner_scale, b * 2 * inner_scale),
            math.degrees(theta),
        )

        # Generate the transparency layer -- the outer ellipse filled and anti-aliased
        transparency = numpy.zeros((height, width), numpy.uint8)
        cv2.ellipse(
            transparency, ellipse_outer, 255, -1, LineTypeEnum.anti_aliased.value
        )
        if save_intermediate:
            show_image(
                transparency,
                wait=True
                # save_path = basep/"eligrad-t.png"
            )

        # Generate the gradient and scale it to 8bit grayscale range
        intensity = numpy.uint8(
            make_gradient_v1(width, height, h, k, a, b, theta) * 255
        )
        if save_intermediate:
            show_image(
                intensity,
                wait=True
                # save_path =  str(basep / "eligrad-i1.png")
            )

        # Draw the inter ellipse filled and anti-aliased
        cv2.ellipse(intensity, ellipse_inner, 255, -1, LineTypeEnum.anti_aliased.value)
        if save_intermediate:
            show_image(
                intensity,
                # save_path =  str(basep / "eligrad-i2.png")
                wait=True,
            )

        # Turn it into a BGRA image
        result = cv2.merge([intensity, intensity, intensity, transparency])
        return result

    # ============================================================================

    a, b = (360.0, 200.0)  # Semi-major and semi-minor axis
    theta = math.radians(40.0)  # Ellipse rotation (radians)
    inner_scale = 0.6  # Scale of the inner full-white ellipse

    show_image(
        draw_image(a, b, theta, inner_scale, True),
        wait=True
        # save_path = str(basep/"eligrad.png")
    )

    # ============================================================================

    rows = []
    for j in range(0, 4, 1):
        cols = []
        for i in range(0, 90, 10):
            tile = numpy.zeros((170, 170, 4), numpy.uint8)
            image = draw_image(80.0, 50.0, math.radians(i + j * 90), 0.6)
            tile[: image.shape[0], : image.shape[1]] = image
            cols.append(tile)
        rows.append(numpy.hstack(cols))

    show_image(
        numpy.vstack(rows),
        # save_path = str(basep/"eligrad-m.png")
        wait=True,
    )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 23/07/2020
           """

from typing import Sequence, Tuple, Union

from draugr import replicate

__all__ = ["pad2d_hw_shape"]


def pad2d_hw_shape(
    h_w: Union[int, Tuple[int, int]],
    pad_size: Union[int, Tuple[int, int], Tuple[int, int, int, int]] = 1,
) -> Tuple[int, int]:
    """

    quad(left,right,top,bottom) torch quad definition
    double(h,w)
    single(all)

      :param pad_size:
    :param h_w:
    :type h_w:
    :return:
    :rtype:"""
    h_w = replicate(h_w)
    if isinstance(pad_size, Sequence) and len(pad_size) == 4:
        pad = (pad_size[2:], pad_size[:2])  # NOTE: permuted torch quadruple pad order
    else:
        pad_size = replicate(pad_size)
        pad = (replicate(pad_size[0]), replicate(pad_size[1]))

    h = h_w[0] + sum(pad[0])  # Sum across pad[0] (top,bottom)
    w = h_w[1] + sum(pad[1])  # Sum across pad[1] (left,right)

    return h, w


if __name__ == "__main__":
    print(pad2d_hw_shape(5, 1))
    print(pad2d_hw_shape(5, (10, 10)))
    print(pad2d_hw_shape(5, (10, 10, 3, 2)))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from abc import abstractmethod
from typing import Any, Sequence, Tuple

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 21/09/2019
           """

__all__ = ["MplDrawer", "MockDrawer"]

import matplotlib
from draugr.drawers.drawer import Drawer
from matplotlib import pyplot


class MplDrawer(
    # metaclass=PostInit
    Drawer
):
    """
    Real time plotting base class

    for Matplotlib"""

    # @drop_unused_kws
    def __init__(
        self,
        *,
        default_delta: float = 1 / 120,
        render: bool = True,
        placement: Tuple[int, int] = None,
        **kwargs,
    ):
        """

        :param default_delta:
        :param render:
        :param placement:
        :param kwargs:"""
        super().__init__(**kwargs)
        self.fig = None

        if not render:
            return

        if default_delta is None:  # Zero still passes
            default_delta = 1 / 120

        self._default_delta = default_delta
        self.n = 0

        """
fig_manager = pyplot.get_current_fig_manager()
geom = fig_manager.window.geometry()
x, y, dx, dy = geom.getRect()
fig_manager.window.setGeometry(*placement, dx, dy)
fig_manager.window.SetPosition((500, 0))
"""
        self.placement = placement

    """
@drop_unused_kws
def __post_init__(self,*, figure_size: Tuple[int, int] = None):
if self.fig is None:
if figure_size is None:
figure_size = (4, 4)
self.fig = pyplot.figure(figsize=figure_size)
"""

    def draw(self, data: Any, delta: float = None):
        """ """
        if not self.fig:
            raise NotImplementedError(
                "Figure was not instantiated check specialisation of MplDrawer"
            )

        self._draw(data)

        pyplot.draw()

        if self.n <= 1 and self.placement:
            self.move_figure(self.fig, *self.placement)
        self.n += 1

        if delta:  # TODO: ALLOW FOR ASYNC DRAWING
            pyplot.pause(delta)
        elif self._default_delta:
            pyplot.pause(self._default_delta)

    @staticmethod
    def move_figure(figure: pyplot.Figure, x: int = 0, y: int = 0):
        r"""
        Move figure's upper left corner to pixel (x, y)"""
        backend = matplotlib.get_backend()
        if hasattr(figure.canvas.manager, "window"):
            window = figure.canvas.manager.window
            if backend == "TkAgg":
                window.wm_geometry(f"+{x:d}+{y:d}")
            elif backend == "WXAgg":
                window.SetPosition((x, y))
            else:
                # This works for QT and GTK
                # You can also use window.setGeometry
                window.move(x, y)

    def __enter__(self):
        return self

    def close(self):
        """ """
        if self._verbose:
            print("mlpdrawer close was called")
        if self.fig:
            pyplot.close(self.fig)

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def __del__(self):
        self.close()

    @abstractmethod
    def _draw(self, data: Any) -> None:
        """

        :param data:
        :type data:"""
        raise NotImplementedError


class MockDrawer(MplDrawer):
    """
    Mock for drawer, accepts data but draws nothing"""

    def _draw(self, data: Sequence) -> None:
        """

        :param data:
        :type data:"""
        pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from typing import Sequence, Union

from draugr.drawers.mpl_drawers.mpldrawer import MplDrawer
from draugr.numpy_utilities import recursive_flatten_numpy

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 22/11/2019

           """

from matplotlib import pyplot

import numpy

__all__ = ["SeriesScrollPlot"]

from warg import passes_kws_to


class SeriesScrollPlot(MplDrawer):
    """ """

    @passes_kws_to(MplDrawer.__init__)
    def __init__(
        self,
        window_length: int = None,
        title: str = "",
        time_label: str = "Time",
        data_label: str = "Magnitude",
        reverse: bool = False,
        overwrite: bool = False,
        render: bool = True,
        **kws
    ):
        """

        :param window_length:
        :param title:
        :param time_label:
        :param data_label:
        :param reverse:
        :param overwrite:
        :param placement:
        :param render:"""

        super().__init__(render=render, **kws)
        if not render:
            return

        self.overwrite = overwrite
        self.reverse = reverse
        self.window_length = window_length

        if window_length:
            assert window_length > 3

        self.fig = pyplot.figure(figsize=(4, 4))
        self.im = pyplot.plot([], [])

        if window_length:
            for i in range(len(self.im)):
                self.im[i].set_xdata(range(window_length))
                self.im[i].set_ydata([numpy.nan] * window_length)
            pyplot.xlim(0, window_length)

        pyplot.xlabel(time_label)
        pyplot.ylabel(data_label)

        pyplot.title(title)
        pyplot.tight_layout()

    def _draw(self, data: Union[Sequence, int, float, complex]):
        """
        SHOULD NOT BE CALLED DIRECTLY!

        :param data:
        :return:"""
        if not isinstance(data, Sequence):
            data = [data]

        data = recursive_flatten_numpy(data)

        num_images = len(self.im)
        num_series = len(data)

        min_min, max_max = None, None

        if num_images != num_series:
            # print('Reinstantiating figures')
            self.im = pyplot.plot(*[[] for _ in range(num_series)] * 2)
            if self.window_length:
                for i in range(len(self.im)):
                    self.im[i].set_xdata(range(self.window_length))
                    self.im[i].set_ydata([numpy.nan] * self.window_length)

        for i in range(num_images):
            time_points = self.im[i].get_xdata()
            data_points = self.im[i].get_ydata()

            if self.window_length:
                if not self.overwrite:
                    if not self.reverse:
                        data_points = numpy.concatenate(
                            (data_points[1:], [data[i]]), axis=0
                        )
                    else:
                        data_points = numpy.concatenate(
                            ([data[i]], data_points[:-1]), axis=0
                        )
                else:
                    if not self.reverse:
                        data_points[self.n % self.window_length] = data[i]
                    else:
                        data_points[
                            self.window_length - 1 - self.n % self.window_length
                        ] = data[i]
            else:
                if not self.reverse:
                    time_points = numpy.concatenate((time_points, [self.n]), axis=0)
                    data_points = numpy.concatenate((data_points, [data[i]]), axis=0)
                else:
                    time_points = numpy.concatenate(([self.n], time_points), axis=0)
                    data_points = numpy.concatenate(([data[i]], data_points), axis=0)
                pyplot.xlim(0, self.n + 1)

            self.im[i].set_xdata(time_points)
            self.im[i].set_ydata(data_points)
            min_, max_ = numpy.nanmin(data_points), numpy.nanmax(data_points)
            if min_min is None or min_ < min_min:
                min_min = min_
            if max_max is None or max_ > max_max:
                max_max = max_

        if min_min == max_max:
            max_max += 1

        pyplot.ylim(min_min, max_max)


if __name__ == "__main__":

    def asidjas() -> None:
        """
        :rtype: None
        """
        from math import cos, sin

        def multi_series():
            """ """
            s = SeriesScrollPlot(
                window_length=200, reverse=False, overwrite=False, default_delta=None
            )
            for i in range(1000):
                s.draw([sin(i / 100) * 2, cos(i / 10)])

        def single_series():
            """ """
            s = SeriesScrollPlot(window_length=200, reverse=False, overwrite=False)
            for i in range(1000):
                s.draw([sin(i / 20)], 1.0 / 60.0)

        def single_series_no_wrap():
            """ """
            s = SeriesScrollPlot(window_length=200, reverse=True, overwrite=False)
            for i in range(1000):
                s.draw(sin(i / 20), 1.0 / 60.0)

        def single_series_no_wrap_rescale():
            """ """
            s = SeriesScrollPlot(window_length=100, reverse=True, overwrite=False)
            for i in range(1000):
                s.draw(sin(i / 100), 1.0 / 60.0)

        multi_series()

    asidjas()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 09/10/2019
           """

from .channel_transform import *
from .conversion import *



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 07-05-2021
           """

import os
from pathlib import Path

__all__ = ["latest_file"]


def latest_file(
    directory: Path,
    extension: str = "",
    *,
    recurse: bool = False,
    raise_on_failure: bool = True,
) -> Path:
    """

    :param directory:
    :param extension:
    :param recurse:
    :param raise_on_failure:
    :return:
    """
    a = f"*{extension}"
    if recurse:
        path_gen = directory.rglob(a)
    else:
        path_gen = directory.glob(a)
    list_of_files = list(path_gen)
    if len(list_of_files) == 0:
        msg = f"Found no previous files with extension {extension} in {directory}"
        if raise_on_failure:
            raise FileNotFoundError(msg)
        print(f"{msg}, returning None!")
        return None
    return max(list_of_files, key=os.path.getctime)  # USES CREATION TIME


if __name__ == "__main__":
    print(latest_file(Path(__file__).parent, recurse=True))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 20/07/2020
           """

import datetime
import os
from typing import Optional, Tuple, Union

import torch
from torch.nn.modules.module import Module
from torch.optim import Optimizer

from draugr.torch_utilities.persistence.config import (
    ensure_directory_exist,
    save_config,
)
from warg.decorators.kw_passing import drop_unused_kws

parameter_extension = ".parameters"
config_extension = ".py"
optimiser_extension = ".optimiser"

__all__ = [
    "load_model_parameters",
    "load_latest_model_parameters",
    "save_parameters_and_configuration",
    "save_model_parameters",
]

from pathlib import Path


@drop_unused_kws
def load_latest_model_parameters(
    model: torch.nn.Module,
    *,
    optimiser: Optimizer = None,
    model_name: str,
    model_directory: Path,
) -> Tuple[Union[torch.nn.Module, Tuple[torch.nn.Module, Optimizer]], bool]:
    """

    inplace but returns model

    :param optimiser:
    :param model:
    :type model:
    :param model_directory:
    :param model_name:
    :return:"""
    if model:
        model_path = model_directory / model_name
        list_of_files = list(model_path.glob(f"*{parameter_extension}"))
        if len(list_of_files) == 0:
            print(
                f"Found no previous models with extension {parameter_extension} in {model_path}"
            )
        else:
            latest_model_parameter_file = max(list_of_files, key=os.path.getctime)
            print(f"loading previous model parameters: {latest_model_parameter_file}")

            model.load_state_dict(torch.load(str(latest_model_parameter_file)))

            if optimiser:
                opt_st_d_file = latest_model_parameter_file.with_suffix(
                    optimiser_extension
                )
                if opt_st_d_file.exists():
                    optimiser.load_state_dict(torch.load(str(opt_st_d_file)))
                    print(f"loading previous optimiser state: {opt_st_d_file}")
                return (model, optimiser), True
            else:
                return model, True
    if optimiser:
        return (model, optimiser), False
    return model, False


load_model_parameters = load_latest_model_parameters


# @passes_kws_to(save_config)
def save_parameters_and_configuration(
    *,
    model: Module,
    model_save_path: Path,
    optimiser: Optional[Optimizer] = None,
    optimiser_save_path: Optional[Path] = None,
    config_save_path: Optional[Path] = None,
    loaded_config_file_path: Optional[Path] = None,
) -> None:
    """

    :param optimiser:
    :type optimiser:
    :param optimiser_save_path:
    :type optimiser_save_path:
    :param model:
    :param model_save_path:
    :param config_save_path:
    :param loaded_config_file_path:
    :return:"""
    torch.save(model.state_dict(), str(model_save_path))
    if optimiser:
        torch.save(optimiser.state_dict(), str(optimiser_save_path))
    if loaded_config_file_path:
        save_config(config_save_path, loaded_config_file_path)


@drop_unused_kws
def save_model_parameters(
    model: Module,
    *,
    model_name: str,
    save_directory: Path,
    optimiser: Optional[Optimizer] = None,
    config_file_path: Optional[Path] = None,
) -> None:
    """

    :param optimiser:
    :param model:
    :param save_directory:
    :param config_file_path:
    :param model_name:
    :return:"""
    model_date = datetime.datetime.now()

    model_time_rep = model_date.strftime("%Y%m%d%H%M%S")
    model_save_path = save_directory / model_name / f"{model_time_rep}"
    ensure_directory_exist(model_save_path.parent)

    saved = False
    try:
        save_parameters_and_configuration(
            model=model,
            model_save_path=model_save_path.with_suffix(parameter_extension),
            optimiser=optimiser,
            optimiser_save_path=(
                model_save_path.parent / f"{model_time_rep}"
            ).with_suffix(optimiser_extension),
            loaded_config_file_path=config_file_path,
            config_save_path=(model_save_path.parent / f"{model_time_rep}").with_suffix(
                config_extension
            ),
        )
        saved = True
    except FileNotFoundError as e:
        print(e)
        while not saved:
            model_save_path = (
                Path(input("Enter another file path: ")).expanduser().resolve()
            )
            ensure_directory_exist(model_save_path.parent)
            try:
                save_parameters_and_configuration(
                    model=model,
                    model_save_path=model_save_path.endswith(parameter_extension),
                    optimiser=optimiser,
                    optimiser_save_path=(
                        model_save_path.parent / f"{model_time_rep}"
                    ).with_suffix(optimiser_extension),
                    loaded_config_file_path=config_file_path,
                    config_save_path=(
                        model_save_path.parent / f"{model_time_rep}"
                    ).with_suffix(config_extension),
                )
                saved = True
            except FileNotFoundError as e:
                print(e)
                saved = False

    if saved:
        print(
            f"Successfully saved model parameters, optimiser state and configuration at names {[model_save_path.with_suffix(a) for a in (parameter_extension, optimiser_extension, config_extension)]}"
        )
    else:
        print(f"Was unsuccessful at saving model or configuration")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 13/11/2019
           """

from typing import Dict, List

import torch
from torch import nn
from torch.nn import Module

__all__ = [
    "set_all_parameter_requires_grad",
    "set_first_n_parameter_requires_grad",
    "trainable_parameters",
    "named_trainable_parameters",
    "trainable_parameters_iterator",
    "named_trainable_parameters_iterator",
]


def set_all_parameter_requires_grad(model: nn.Module, bo: bool = False) -> None:
    """

    :param model:
    :type model:
    :param bo:
    :type bo:"""
    for param in model.parameters():
        param.requires_grad = bo


def set_first_n_parameter_requires_grad(
    model: nn.Module, n: int = 6, bo: bool = False
) -> None:
    """

    :param model:
    :type model:
    :param n:
    :type n:
    :param bo:
    :type bo:"""
    for i, child in enumerate(model.children()):
        if i <= n:
            set_all_parameter_requires_grad(child, bo)


def trainable_parameters(model: Module) -> List[torch.nn.Parameter]:
    """

    :param model:
    :type model:
    :return:
    :rtype:"""
    params_to_update = []
    for name, param in model.named_parameters():
        if param.requires_grad:
            params_to_update.append(param)
    return params_to_update


def named_trainable_parameters(model: Module) -> Dict[str, torch.nn.Parameter]:
    """

    :param model:
    :type model:
    :return:
    :rtype:"""
    params_to_update = {}
    for name, param in model.named_parameters():
        if param.requires_grad:
            params_to_update[name] = param
    return params_to_update


def trainable_parameters_iterator(model: Module) -> List[torch.nn.Parameter]:
    """ """
    for name, param in model.named_parameters():
        if param.requires_grad:
            yield param


def named_trainable_parameters_iterator(model: Module) -> Dict[str, torch.nn.Parameter]:
    """

    :param model:
    :type model:
    :return:
    :rtype:"""
    for name, param in model.named_parameters():
        if param.requires_grad:
            yield {name: param}



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from abc import ABC, abstractmethod

import torch

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 09/10/2019
           """
__all__ = ["GraphWriterMixin"]


class GraphWriterMixin(ABC):
    """
    Writer mixin that provides an interface for 'writing' graphs"""

    @abstractmethod
    def graph(
        self, model: torch.nn.Module, input_to_model: torch.Tensor, **kwargs
    ) -> None:
        """

            :param model:
            :param input_to_model:
        :param kwargs:
        :type kwargs:"""
        raise NotImplementedError



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 28/06/2020
           """

import itertools
import sys

import torch

__all__ = ["size_of_tensor"]


def size_of_tensor(obj: torch.Tensor) -> int:
    r"""**Get size in bytes of Tensor, torch.nn.Module or standard object.**

    Specific routines are defined for torch.tensor objects and torch.nn.Module
    objects. They will calculate how much memory in bytes those object consume.

    If another object is passed, `sys.getsizeof` will be called on it.

    This function works similarly to C++'s sizeof operator.


    Parameters
    ----------
    obj
      Object whose size will be measured.

    Returns
    -------
    int
      Size in bytes of the object"""
    if torch.is_tensor(obj):
        return obj.element_size() * obj.numel()

    elif isinstance(obj, torch.nn.Module):
        return sum(
            size_of_tensor(tensor)
            for tensor in itertools.chain(obj.buffers(), obj.parameters())
        )
    else:
        return sys.getsizeof(obj)


if __name__ == "__main__":
    module = torch.nn.Linear(20, 20)
    bias = 20 * 4  # in bytes
    weights = 20 * 20 * 4  # in bytes
    assert size_of_tensor(module) == bias + weights



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 11/05/2020
           """

import torch
from torch.nn import Module

from draugr.torch_utilities import global_torch_device
from warg import AlsoDecorator

__all__ = ["TorchCpuSession", "TorchCudaSession", "TorchDeviceSession"]


class TorchCudaSession(AlsoDecorator):
    """
    Sets global torch devices to cuda if available"""

    def __init__(self, model: Module = None, no_side_effect: bool = True):
        self._model = model
        self._no_side_effect = no_side_effect
        if no_side_effect:
            self.prev_dev = global_torch_device()

    def __enter__(self):
        device = global_torch_device(override=torch.device("cuda"))
        if self._model:
            self._model.to(device)

        return True

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self._no_side_effect:
            device = global_torch_device(override=self.prev_dev)
        else:
            device = global_torch_device(override=torch.device("cpu"))
        if self._model:
            self._model.to(device)
        return False


class TorchCpuSession(AlsoDecorator):
    """
    Sets global torch devices to cpu"""

    def __init__(self, model: Module = None, no_side_effect: bool = True):
        self._model = model
        self._no_side_effect = no_side_effect
        if no_side_effect:
            self.prev_dev = global_torch_device()

    def __enter__(self):
        device = global_torch_device(override=torch.device("cpu"))
        if self._model:
            self._model.to(device)
        return True

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self._no_side_effect:
            device = global_torch_device(override=self.prev_dev)
        else:
            device = global_torch_device(override=torch.device("cuda"))
        if self._model:
            self._model.to(device)
        return False


class TorchDeviceSession(AlsoDecorator):
    """
    Sets global torch devices to cpu"""

    def __init__(
        self, device: torch.device, model: Module = None, no_side_effect: bool = True
    ):
        self._model = model
        self._no_side_effect = no_side_effect
        self._device = device
        if no_side_effect:
            self.prev_dev = global_torch_device()

    def __enter__(self):
        device = global_torch_device(override=self._device)
        if self._model:
            self._model.to(device)
        return True

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self._no_side_effect:
            device = global_torch_device(override=self.prev_dev)
            if self._model:
                self._model.to(device)
        return False


if __name__ == "__main__":
    with TorchDeviceSession(global_torch_device()):
        pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 20/10/2019
           """

__all__ = ["set_lr"]


def set_lr(optimizer, lr):
    """

    :param optimizer:
    :type optimizer:
    :param lr:
    :type lr:"""
    for param_group in optimizer.param_groups:
        param_group["lr"] = lr



__all__ = [
    "get_screen_resolution",
    "screen_res_tk",
    "screen_res_pygame",
    "screen_res_gtk",
    "screen_res_xlib",
    "screen_res_mac",
    "screen_res_win",
]

import sys


def screen_res_xlib():
    import xlib

    root = xlib.display.Display().screen().root
    return root.get_geometry()[2], root.get_geometry()[3]


def screen_res_pygame():
    import pygame

    return pygame.display.Info().current_w, pygame.display.Info().current_h


def screen_res_gtk():
    import gtk

    screen = gtk.Window().get_screen()
    return screen.get_width(), screen.get_height()


def screen_res_tk():
    import Tkinter

    root = Tkinter.Tk()
    screen_width = root.winfo_screenwidth()
    screen_height = root.winfo_screenheight()
    return screen_width, screen_height


def screen_res_win():
    import ctypes.wintypes

    return ctypes.windll.user32.GetSystemMetrics(
        0
    ), ctypes.windll.user32.GetSystemMetrics(1)


def screen_res_mac():
    try:
        import Quartz
    except:
        assert (
            False
        ), "You must first install pyobjc-core and pyobjc: MACOSX_DEPLOYMENT_TARGET=10.11 pip install pyobjc"
    main_disp_id = Quartz.CGMainDisplayID()
    return Quartz.CGDisplayPixelsWide(main_disp_id), Quartz.CGDisplayPixelsHigh(
        main_disp_id
    )


def get_screen_resolution():
    if sys.platform == "win32":
        return screen_res_win()
    elif sys.platform == "darwin":
        return screen_res_mac()
    elif sys.platform == "linux":
        return screen_res_gtk()
    elif sys.platform == "linux2":
        return screen_res_gtk()
    elif sys.platform == "cygwin":
        return screen_res_gtk()
    else:
        return screen_res_tk()



if __name__ == "__main__":

    def aisjhd():

        from matplotlib import pyplot
        import cv2
        from draugr.opencv_utilities import LineTypeEnum
        import numpy
        import random

        ANGLE_DELTA = 360 // 8

        img = numpy.zeros((700, 700, 3), numpy.uint8)
        img[::] = 255

        for size in range(300, 0, -100):
            for angle in range(0, 360, ANGLE_DELTA):
                r = random.randint(0, 256)
                g = random.randint(0, 256)
                b = random.randint(0, 256)
                cv2.ellipse(
                    img,
                    (350, 350),
                    (size, size),
                    0,
                    angle,
                    angle + ANGLE_DELTA,
                    (r, g, b),
                    LineTypeEnum.filled.value,
                )

        pyplot.gcf().set_size_inches((8, 8))
        pyplot.imshow(img)
        pyplot.show()

    aisjhd()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 18-02-2021
           """

__all__ = [
    "monochrome_hatch_cycler",
    "simple_hatch_cycler",
    "monochrome_line_no_marker_cycler",
    "monochrome_line_cycler",
]

from matplotlib import cycler

from draugr.visualisation.matplotlib_utilities.styles.hatching import (
    four_times_denser_hatch,
)
from draugr.visualisation.matplotlib_utilities.styles.lines import (
    line_styles,
    marker_styles,
)

color_cycler = cycler("color", ["E24A33", "348ABD", "988ED5", "FBC15E", "8EBA42"])
marker_cycler = cycler("marker", marker_styles)
line_cycler = cycler("linestyle", line_styles)
simple_hatch_cycler = cycler("hatch", four_times_denser_hatch)

monochrome_hatch_cycler = (
    cycler("color", "w")
    * cycler("facecolor", "w")
    * cycler("edgecolor", "k")
    * simple_hatch_cycler
)

monochrome_line_no_marker_cycler = cycler("color", ["k"]) * cycler(
    "linestyle", line_styles
)

monochrome_line_cycler = cycler("color", ["k"]) * line_cycler * marker_cycler

if __name__ == "__main__":
    print([a for _, a in zip(range(10), monochrome_line_cycler)])



import os
from pathlib import Path
from typing import Union, Iterable

os.environ["OPENCV_IO_ENABLE_OPENEXR"] = "1"
import cv2
import numpy

from apppath import ensure_existence
from draugr.opencv_utilities.namespaces.enums import WindowFlagEnum, FontEnum
from draugr.opencv_utilities.windows.default import (
    ESC_CHAR,
    ExtensionEnum,
    match_return_code,
)
from warg import get_first_arg_name

__all__ = ["show_image"]


def show_image(
    image: numpy.ndarray,
    title: str = None,
    *,
    flag: WindowFlagEnum = WindowFlagEnum.gui_expanded.value,
    wait: Union[bool, int] = False,
    draw_title: bool = False,
    save_path: Path = None,
    exit_chars: Iterable[str] = ("q", ESC_CHAR),
    extension: ExtensionEnum = ExtensionEnum.exr,  # 'png'
    min_default_size=200,
    max_default_size=600,
) -> None:
    """
    ! if a title is not provided ( None) , title will be inferred. Caution in real time imshow / animations this will hurt performance.

    :param image:
    :type image:
    :param title:
    :type title:
    :param flag:
    :type flag:
    :param wait:
    :type wait:
    :param draw_title:
    :type draw_title:
    :param save_path:
    :type save_path:
    :param exit_chars:
    :type exit_chars:
    :param extension:
    :type extension:
    :return:
    :rtype:
    """
    if title is None:
        title = get_first_arg_name("show_image", verbose=True)
        if title is None:
            title = "image"

    if draw_title:
        cv2.putText(
            image.copy(),
            title,
            (25, image.shape[0] - 20),
            FontEnum.hershey_simplex.value,
            2.0,
            (0, 255, 0),
            3,
        )
    try:
        cv2.getWindowProperty(title, 0)  # is open test
    except:
        cv2.namedWindow(title, flag)

        w_o, h_o = image.shape[1], image.shape[0]
        max_d = max(
            max(min(w_o, max_default_size), min(h_o, max_default_size)),
            min_default_size,
        )
        ar_o = h_o / w_o
        w, h = max_d, ar_o * max_d

        cv2.resizeWindow(title, int(w), int(h))

    cv2.imshow(title, image)

    if save_path is not None:
        extension = ExtensionEnum(extension)
        if extension == ExtensionEnum.exr:
            assert (
                os.environ.get("OPENCV_IO_ENABLE_OPENEXR") == "1"
            ), f'Openexr is support not enabled, must be declared before import of cv2, OPENCV_IO_ENABLE_OPENEXR{os.environ.get("OPENCV_IO_ENABLE_OPENEXR")}'
            image = image.astype(dtype=numpy.float32)
            # cv2.

        ensure_existence(save_path.parent)

        cv2.imwrite(str(save_path.with_suffix(f".{extension.value}")), image)

    if wait is not None:
        """# WEIRDO PYTHON
        if (
            wait is int or wait is float and wait >= 0
        ):  # DO NOT REFACTOR TO ISINSTANCE as bool is an instance of int!
          return ret_val_comp(cv2.waitKey(wait), char)
        else:
          return ret_val_comp(cv2.waitKey(), char)
        """
        if isinstance(wait, bool):
            if wait:
                return match_return_code(cv2.waitKey(), exit_chars)
        elif wait >= 0:
            return match_return_code(cv2.waitKey(wait), exit_chars)
    return False


if __name__ == "__main__":

    def assd():
        aasdsad = numpy.zeros((50, 50))
        print(show_image(aasdsad, wait=True))
        asd_a_as = numpy.zeros((50, 50))
        print(show_image(asd_a_as, wait=True, save_path=Path("exclude") / "out.exr"))

    assd()
